# RCAIDE/Library/Methods/Powertrain/Converters/Generator/compute_generator_performance.py

# 
# Created: Feb 2025, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
import RCAIDE

# package imports 
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  compute_omega_and_Q_from_Cp_and_V
# ----------------------------------------------------------------------------------------------------------------------    
def compute_generator_performance(generator,conditions):
    """
    Computes generator performance characteristics including electrical, mechanical and thermal parameters.

    Parameters
    ----------
    generator : Converter
        Generator component for which performance is being computed
    generator_conditions : Conditions
        Container for generator operating conditions
    conditions : Conditions 
        Mission segment conditions containing freestream properties

    Returns
    -------
    None
        Updates generator_conditions in-place with computed performance parameters

    Notes
    -----
    This function handles both PMSM and DC generator types with different computation approaches:
        - Uses speed-torque relationships
        - Accounts for gearbox effects
        - Computes electrical parameters (current, voltage)
        - Determines overall efficiency

    **Major Assumptions**
        * Steady state operation
        * Uniform temperature distribution
        * No magnetic saturation effects
        * Linear speed-torque characteristics for DC generators
        * Constant material properties

    See Also
    --------
    RCAIDE.Library.Components.Powertrain.Converters.DC_Generator
    RCAIDE.Library.Components.Powertrain.Converters.PMSM_Generator
    """
    
    # unpack generator conditions 
    generator_conditions = conditions.energy.converters[generator.tag]    
 
    if type(generator) == RCAIDE.Library.Components.Powertrain.Converters.DC_Generator:   
        if generator.inverse_calculation == False:
            power          = generator_conditions.inputs.power 
            Res            = generator.resistance  
            Kv             = generator.speed_constant
            G              = generator.gearbox.gear_ratio 
            io             = generator.no_load_current  
            v              = generator_conditions.outputs.voltage 
            omega          = generator_conditions.inputs.omega
            omega_internal = omega * G
            i              = (v - (omega_internal) /Kv)/Res  
            Q              = power / omega  
            etam           = (1-io/i)*(1-i*Res/v)
            
        else:
            Res             = generator.resistance  
            Kv              = generator.speed_constant
            G               = generator.gearbox.gear_ratio 
            io              = generator.no_load_current  
            v               = generator_conditions.outputs.voltage             
            i               = generator_conditions.outputs.current
            P_elec          = i * v
            omega_internal  = ((v - (Res * i)) * Kv)   
            Q_internal      = (((v-omega_internal /Kv)/Res -io)/Kv)
            omega           = omega_internal / G
            Q               = Q_internal * G
            etam            = (1-io/i)*(1-i*Res/v)
        
        generator_conditions.outputs.current    = i 
        generator_conditions.outputs.power      = i *v  
        generator_conditions.inputs.power       = Q * omega 
        generator_conditions.inputs.torque     = Q
        generator_conditions.inputs.omega      = omega
        generator_conditions.inputs.efficiency = etam          
        
    elif type(generator) == RCAIDE.Library.Components.Powertrain.Converters.PMSM_Generator: 
        if generator.inverse_calculation == False:
            io     = generator.no_load_current
            G      = generator.gearbox.gear_ratio 
            omega  = generator_conditions.inputs.omega  
            power  = generator_conditions.inputs.power  
            Kv     = generator.speed_constant                  
            D_in   = generator.inner_diameter         
            kw     = generator.winding_factor     
            Res    = generator.resistance                      
            L      = generator.stack_length                    
            l      = generator.length_of_path                  
            mu_0   = generator.mu_0                            
            mu_r   = generator.mu_r   
            Q      = power/omega                               
            i      = np.sqrt((2*(Q/G)*l)/(D_in*mu_0*mu_r*L*kw))           
            v      = (omega * G)/((2 * np.pi / 60)*Kv) + i*Res        
            etam   = (1-io/i)*(1-i*Res/v) 
        
            generator_conditions.outputs.current    = i 
            generator_conditions.outputs.power      = i *v    
            generator_conditions.outputs.efficiency = etam   
        else:
            Res            = generator.resistance
            G              = generator.gearbox.gear_ratio
            I              = generator_conditions.outputs.current
            V              = generator_conditions.outputs.voltage
            I_turn         = I/generator.number_of_turns                                                             # [A]            current in each turn
            omega          = (generator.speed_constant*(V - I*Res)) /G                                     # [RPM -> rad/s] rotor angular velocity
            A              = np.pi * ((generator.stator_outer_diameter**2 - generator.stator_inner_diameter**2) / 4)     # [m**2]         cross-sectional area of the reluctance path perpendicular to length 𝑙    
            MMF_coil       = generator.number_of_turns*I_turn                                                        # [A*turns]      magnetomotive force applied to the reluctance path for a coil (Eq.14)  
            R              = generator.length_of_path/(A*generator.mu_0*generator.mu_r)                                      # [A*turn/Wb]    reluctance of a given path or given reluctant element (Eq.16) 
            phi            = MMF_coil/R                                                                          # [Wb]           magnetic flux through the reluctance path (Eq.12)
            B_sign         = phi/A                                                                               # [V*s/m**2]     ranges from 0.5 to 1.2, average magnitude of the radial flux density produced by the rotor
            A_sign         = (generator.winding_factor*I)/(np.pi*generator.stator_inner_diameter)                        # [-]            stator electrical loading (Eq.2)        
            TQ             = (np.pi/2)*(B_sign*A_sign)*(generator.inner_diameter**2)*generator.stack_length # [Nm]           torque (Eq.1)
            P              = omega*TQ                                                                            # [W]            power (Eq.1)        
            A              = np.pi * ((generator.stator_outer_diameter**2 - generator.stator_inner_diameter**2) / 4)     # [m**2]         cross-sectional area of the reluctance path perpendicular to length 𝑙    
            
            generator_conditions.inputs.torque           = TQ 
            generator_conditions.inputs.omega            = omega 
            generator_conditions.inputs.power            = P

   
 
    return


# RCAIDE/Library/Methods/Mass_Properties/Weight_Buildups/Conventional/Transport/Common/__init__.py
# 

"""
Methods for computing component weights that are common across different aircraft configurations. 
This module provides weight estimation techniques for standard aircraft components and systems 
that share similar design principles regardless of aircraft type.

See Also
--------
RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.Transport
    Transport-specific weight methods
RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.General_Aviation
    General aviation weight methods
RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.BWB
    BWB-specific weight methods
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from .compute_payload_weight import compute_payload_weight



# RCAIDE/Library/Attributes/Solids/Polytetrafluoroethylene.py
# 

# Created: Jan 2025 M. Clarke

#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------

from .Solid import Solid 
from array import * 

#-------------------------------------------------------------------------------
# Polytetrafluoroethylene Insulation Material
#------------------------------------------------------------------------------- 
class Polytetrafluoroethylene(Solid): 
    """
    A class representing Polytetrafluoroethylene (PFTE) and its material properties.

    Attributes
    ----------
    electrical_permittivity : float
        Material electrical permittivity in kg/m³ (2.1)
    dielectric_strength_range : list
        Range of dielectric strength in Pa (6E7,8E7)
    density : float
        Material density in kg/m³ (2170)
    thermal_conductivity : float
        Heat conduction coefficient in W/(m·K) (0.25)
    melting_point : float
        Material melting point in K (600)
    temperature_range : list
        Range of temperature in K (183, 533)
    modulus_of_elasticity : float
        Material modulus of elasticity in Pa (.49E9)
    yield_tensile_strength : float
        Stress at which material begins to deform plastically in Pa (24e6)
    """

    def __defaults__(self):
        """Sets material properties at instantiation. 

        Assumptions:
            None
    
        Source:
            Guo, Ruochen, et al. "Electrical Architecture of 90-seater Electric Aircraft: A Cable Perspective."
            IEEE Transactions on Transportation Electrification (2024).
        """
        self.electrical_permittivity    = 2.1
        self.dielectric_strength_range  = [6E7, 8E7] # [V/m]
        self.density                    = 2170
        self.thermal_conductivity       = 0.25
        self.melting_point              = 600 
        self.temperature_range          = [183 , 533]  
        self.modulus_of_elasticity      = 0.49E9
        self.yield_tensile_strength     = 24E6
        return 


# RCAIDE/Library/Compoments/Component.py
#  
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

from RCAIDE.Framework.Core import Container as ContainerBase
from RCAIDE.Framework.Core import Data
from .Mass_Properties import Mass_Properties

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Component
# ----------------------------------------------------------------------------------------------------------------------         
class Component(Data):
    """
    Base class for all physical components in an aircraft configuration.

    Attributes
    ----------
    tag : str
        Unique identifier for the component, defaults to 'Component'
        
    mass_properties : Mass_Properties
        Mass and inertia properties, initialized empty
        
    origin : ndarray
        3D coordinates [x, y, z] defining component's reference point, 
        defaults to [0.0, 0.0, 0.0]

    Notes
    -----
    The Component class serves as the foundation for all physical parts in RCAIDE. 
    It provides:
    
    * Basic geometric positioning
    * Mass properties tracking
    * Container functionality for sub-components

    See Also
    --------
    RCAIDE.Library.Components.Mass_Properties
        Class containing mass and inertia data
    RCAIDE.Framework.Core.Data
        Parent class providing data structure functionality
    """
    def __defaults__(self):
        """
        Sets default values for the component attributes.
        """         
        self.tag             = 'Component' 
        self.mass_properties = Mass_Properties()
        self.origin          = np.array([[0.0,0.0,0.0]])
    
        
# ----------------------------------------------------------------------------------------------------------------------
#  Component Container
# ----------------------------------------------------------------------------------------------------------------------     
class Container(ContainerBase):
    """
    Container class for managing collections of components.

    Notes
    -----
    The Container class provides organization and mass calculation functionality 
    for groups of components. Key features include:
    
    * Recursive mass summation
    * Moment calculation about reference points
    * Component hierarchy management

    See Also
    --------
    RCAIDE.Framework.Core.Container
        Parent class providing base container functionality
    """
    pass 
    
# ------------------------------------------------------------
#  Handle Linking
# ------------------------------------------------------------

Component.Container = Container



# RCAIDE/Library/Missions/Segments/Cruise/Constant_Mach_Constant_Altitude_Loiter.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# Package imports 
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def initialize_conditions(segment):
    """
    Initializes conditions for constant speed loiter at fixed altitude

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed
            - altitude : float
                Loiter altitude [m]
            - time : float
                Loiter duration [s]
            - air_speed : float
                True airspeed to maintain [m/s]
            - sideslip_angle : float
                Aircraft sideslip angle [rad]
            - state:
                numerics.dimensionless.control_points : array
                    Discretization points [-]
                conditions : Data
                    State conditions container
                initials : Data, optional
                    Initial conditions from previous segment
    
    Returns
    -------
    None
        Updates segment conditions directly:
            - conditions.frames.inertial.velocity_vector [m/s]
            - conditions.frames.inertial.position_vector [m]
            - conditions.freestream.altitude [m]
            - conditions.frames.inertial.time [s]


    Notes
    -----
    This function sets up the initial conditions for a loiter segment with constant
    true airspeed and constant altitude. The segment duration is specified by time
    rather than distance.

    **Calculation Process**
        1. Check for initial conditions
        2. Discretize time points over loiter duration
        3. Decompose velocity into components using:
            - Constant airspeed
            - Sideslip angle

    **Major Assumptions**
        * Constant true airspeed
        * Constant altitude
        * Small angle approximations
        * Quasi-steady flight
        * No wind effects

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    """    
    
    # unpack
    alt        = segment.altitude
    final_time = segment.time
    air_speed  = segment.air_speed        
    beta       = segment.sideslip_angle
    conditions = segment.state.conditions 
    
    # check for initial velocity
    if air_speed is None: 
        if not segment.state.initials: raise AttributeError('airspeed not set')
        air_speed = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])
        
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]       
    
    # dimensionalize time
    t_initial = conditions.frames.inertial.time[0,0]
    t_final   = final_time + t_initial
    t_nondim  = segment.state.numerics.dimensionless.control_points
    time      = t_nondim * (t_final-t_initial) + t_initial
    v_x       = np.cos(beta)*air_speed 
    v_y       = np.sin(beta)*air_speed 
    
    # pack
    segment.state.conditions.freestream.altitude[:,0]             = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down
    segment.state.conditions.frames.inertial.velocity_vector[:,0] = v_x
    segment.state.conditions.frames.inertial.velocity_vector[:,1] = v_y
    segment.state.conditions.frames.inertial.time[:,0]            = time[:,0]
    



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Raymer/compute_fuselage_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core    import Units

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
# fuselage Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_fuselage_weight(vehicle, fuselage, settings):
    """
    Calculates the weight of the fuselage for transport aircraft using Raymer's method.

    Parameters
    ----------
    vehicle : RCAIDE.Vehicle()
        Vehicle data structure containing:
            - mass_properties.max_takeoff : float
                Maximum takeoff weight [kg]
            - flight_envelope.ultimate_load : float
                Ultimate load factor (default: 3.75)
            - wings['main_wing'] : Data()
                Main wing properties including taper and sweep
    fuselage : RCAIDE.Component()
        Fuselage component containing:
            - lengths.total : float
                Total fuselage length [m]
            - width : float
                Maximum fuselage width [m]
            - heights.maximum : float
                Maximum fuselage height [m]
    settings : Data()
        Configuration settings containing:
            - Raymer.fuselage_mounted_landing_gear_factor : float
                Factor for fuselage-mounted landing gear
            - Raymer.cargo_doors_number : int
                Number of cargo doors
            - Raymer.cargo_doors_clamshell : bool
                True if cargo doors are clamshell doors, False otherwise

    Returns
    -------
    weight_fuselage : float
        Weight of the fuselage structure [kg]

    Notes
    -----
    This method implements Raymer's semi-empirical correlation for transport aircraft
    fuselage weight estimation. The correlation accounts for size, loads, and wing-body
    intersection effects. It also accounts for the number of cargo doors and their type (if present).

    **Major Assumptions**
        * No fuselage-mounted landing gear by default
        * One cargo door (Kdoor = 1.06)
        * Correlation based on transport category aircraft data

    **Theory**
    The fuselage weight is calculated using:
    .. math::
        W_{fus} = 0.328K_{door}K_{lg}(W_{dg}N_{ult})^{0.5}L^{0.25}S_f^{0.302}(1+K_{ws})^{0.04}(L/D)^{0.1}

    where:
        * :math:`K_{ws}` accounts for wing-body intersection effects
        * :math:`S_f` is the fuselage wetted area
        * :math:`L/D` is the fuselage fineness ratio

    References
    ----------
    [1] Raymer, D., "Aircraft Design: A Conceptual Approach", AIAA 
        Education Series, 2018. 

    See Also
    --------
    RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.Transport.Raymer.compute_main_wing_weight
    RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.Transport.Raymer.compute_operating_empty_weight
    """
    Klg         = settings.fuselage_mounted_landing_gear_factor
    DG          = vehicle.mass_properties.max_takeoff / Units.lbs
    length      = fuselage.lengths.total/ Units.ft
    fuselage_w  = fuselage.width / Units.ft
    fuselage_h  = fuselage.heights.maximum / Units.ft
    
    if settings.cargo_doors_number == 0:
        if settings.cargo_doors_clamshell:
            Kdoor       = 1.12  # clamshell cargo door
        else:
            Kdoor       = 1.  # 0 cargo door
    if settings.cargo_doors_number == 1:
        Kdoor       = 1.06  # 1 cargo door
    if settings.cargo_doors_number == 2:
        if settings.cargo_doors_clamshell:
            Kdoor       = 1.25  # clamshell cargo door
        else:
            Kdoor       = 1.12  # 2 cargo door
   

    D           = (fuselage_w + fuselage_h) / 2.
    Sf          = np.pi * (length/ D - 1.7) * D ** 2  # fuselage wetted area, ft**2
    wing        = vehicle.wings['main_wing']
    Kws         = 0.75 * (1 + 2 * wing.taper) / (1 + wing.taper) * (wing.spans.projected / Units.ft *
                                                            np.tan(wing.sweeps.quarter_chord)) / length

    weight_fuselage = 0.328 * Kdoor * Klg * (DG * vehicle.flight_envelope.ultimate_load) ** 0.5 * length** 0.25 * \
                 Sf ** 0.302 * (1 + Kws) ** 0.04 * (length/ D) ** 0.1
    
    if settings.advanced_composites:
        weight_fuselage *= 0.925

    return weight_fuselage * Units.lbs


# RCAIDE/Methods/Powertrain/Sources/Batteries/Lithium_Ion_NMC/__init__.py
# 

"""
This module provides functionality for modeling lithium-ion NMC (nickel manganese cobalt) batteries in powertrains. 
It includes methods for computing NMC cell performance and updating cell age.

See Also
--------
RCAIDE.Library.Components.Powertrain.Sources.Battery_Modules.Lithium_Ion_NMC
RCAIDE.Library.Methods.Powertrain.Sources.Batteries.Lithium_Ion_LFP
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from .compute_nmc_cell_performance   import compute_nmc_cell_performance, reuse_stored_nmc_cell_data
from .update_nmc_cell_age            import update_nmc_cell_age


# Rotor_Wake provides the functions needed to perform analyses.

"""
Methods for analyzing and computing the performance of rotors in aircraft propulsion systems.

This module provides functions for calculating the aerodynamic performance of various types
of rotors, including propellers, lift rotors, and prop rotors. It implements different
analysis methods such as Blade Element Momentum Theory (BEMT) with Helmholtz wake modeling
and Actuator Disc Theory.

The performance analysis methods calculate key parameters such as thrust, torque, power,
efficiency, and induced velocities based on the rotor geometry, operating conditions, and
flight state. These methods are essential for predicting the behavior of rotors in different
flight regimes, from hover to forward flight.

See Also
--------
RCAIDE.Library.Methods.Powertrain.Converters.Rotor.Design
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# -------------------------------------------------------------------------------------------------------------------- 
from . import Blade_Element_Momentum_Theory_Helmholtz_Wake
from . import Actuator_Disc_Theory









# RCAIDE/Library/Methods/Powertrain/Converters/Ducted_Fan/__init__.py

# Created:  Jan 2025, M. Clarke

"""
Ducted Fan - Blade Element Momentum Theory Methods Package
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .BEMT_performance                   import BEMT_performance
from .purge_files                        import purge_files
from .read_results                       import read_results
from .run_dfdc_analysis                  import run_dfdc_analysis
from .translate_conditions_to_dfdc_cases import translate_conditions_to_dfdc_cases  
from .write_geometry                     import write_geometry
from .write_input_deck                   import write_input_deck 


# RCAIDE/Library/Methods/Powertrain/Converters/Supersonic_Nozzle/append_supersonic_nozzle_conditions.py
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
# append_supersonic_nozzle_conditions 
# ----------------------------------------------------------------------------------------------------------------------    
def append_supersonic_nozzle_conditions(supersonic_nozzle, segment, energy_conditions): 
    """
    Initializes and appends supersonic nozzle conditions data structures to the propulsor conditions dictionary.
    
    Parameters
    ----------
    supersonic_nozzle : SupersonicNozzle
        The supersonic nozzle component for which conditions are being initialized.
    segment : Segment
        The mission segment in which the supersonic nozzle is operating.
    energy_conditions : dict
        Dictionary containing conditions for all propulsion components.
    
    Returns
    -------
    None
        This function modifies the propulsor_conditions dictionary in-place.
    
    Notes
    -----
    This function creates empty Conditions objects for the supersonic nozzle's inputs and outputs
    within the propulsor_conditions dictionary. These conditions will be populated during
    the mission analysis process.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Supersonic_Nozzle.compute_supersonic_nozzle_performance
    """
    energy_conditions.converters[supersonic_nozzle.tag]                     = Conditions()
    energy_conditions.converters[supersonic_nozzle.tag].inputs              = Conditions()
    energy_conditions.converters[supersonic_nozzle.tag].outputs             = Conditions()
    return 


# RCAIDE/Methods/Energy/Propulsors/Converters/Turbofan/__init__.py
# 

"""
Collection of methods for modeling and analyzing turbofan propulsion systems. This module provides functionality for designing, 
sizing, and computing performance characteristics of turbofan engines including thrust calculations and core sizing.
 
See Also
--------
RCAIDE.Library.Components.Powertrain.Propulsors.Turbofan
RCAIDE.Framework.Analysis.Propulsion
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from .append_turbofan_conditions                                 import append_turbofan_conditions 
from .compute_thurst                                             import compute_thrust
from .size_core                                                  import size_core 
from .compute_turbofan_performance                               import compute_turbofan_performance 
from .compute_turbofan_performance                               import reuse_stored_turbofan_data
from .design_turbofan                                            import design_turbofan    



# RCAIDE/Compoments/Propulsors/Converters/Reformer.py
# 
# 
# Created:  Jan 2025, M. Clarke, M. Guidotti

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Core              import Data
from .Converter  import Converter
import numpy as np
import scipy as sp

# ---------------------------------------------------------------------------------------------------------------------- 
#  Nacalle
# ----------------------------------------------------------------------------------------------------------------------  
class Reformer(Converter):
    """
    Reformer Component Class

    This class models a fuel reformer that converts Jet-A fuel into hydrogen-rich reformate gas. 
    It inherits from the base Converter class and implements reformer-specific attributes and methods.

    Attributes
    ----------
    tag : str
        Identifier for the reformer component, defaults to 'reformer'

    x_H : float
        Mass fraction of hydrogen content in Jet-A [-]
    x_C : float 
        Mass fraction of carbon content in Jet-A [-]
    y_H2 : float
        Mole fraction of hydrogen content in reformate [mol]
    y_CO : float
        Mole fraction of carbon monoxide content in reformate [mol]
    rho_F : float
        Density of Jet-A [g/cm^3]
    rho_S : float 
        Density of water [g/cm^3]
    rho_A : float
        Density of air [g/cm^3]
    MW_F : float
        Average molecular weight of Jet-A [g/g-mol]
    MW_S : float
        Average molecular weight of steam [g/g-mol]
    MW_C : float
        Average molecular weight of carbon [g/g-mol]
    MW_H2 : float
        Average molecular weight of hydrogen [g/g-mol]
    A_F_st_Jet_A : float
        Stoichiometric air-to-fuel mass ratio [lb_Air/lb_Jet_A]
    theta : float
        Contact time [sec^-1]
    LHV_F : float
        Lower heating value of Jet-A [kJ/g-mol]
    LHV_H2 : float
        Lower heating value of Hydrogen [kJ/g-mol]
    LHV_CO : float
        Lower heating value of Carbon Monoxide [kJ/g-mol]
    V_cat : float
        Catalyst bed volume [cm^3]
    eta : float
        Reformer efficiency [-]

    Notes
    -----
    The reformer model includes parameters for:
        * Fuel composition and properties
        * Reformate composition
        * Reformer geometry and performance characteristics
        * Thermodynamic properties of reactants/products

    """
    
    def __defaults__(self):
        """ 
        """      
        
        self.tag          = 'reformer'  

        # Jet-A parameters
        self.x_H          = 0.1348   # [-]               mass fraction of hydrogen content in Jet-A
        self.x_C          = 0.8637   # [-]               mass fraction of carbon content in Jet-A
        
        # Reformate parameters
        self.y_H2         = 0.9      # [mol]             mole fraction of hydrogen content in reformate
        self.y_CO         = 0.3      # [mol]             mole fraxtion of carbon monoxide content in reformate
    
        # Reformer parameters
        self.rho_F        = 0.813    # [g/cm**3]         Density of Jet-A
        self.rho_S        = 1        # [g/cm**3]         Density of water
        self.rho_A        = 0.001293 # [g/cm**3]         Density of air
        self.MW_F         = 160      # [g/g-mol]         Average molecular weight of Jet-A    
        self.MW_S         = 18.01    # [g/g-mol]         Average molecular weight of steam
        self.MW_C         = 12.01    # [g/g-mol]         Average molecular weight of carbon
        self.MW_H2        = 2.016    # [g/g-mol]         Average molecular weight of hydrogen
        self.A_F_st_Jet_A = 14.62    # [lb_Air/lb_Jet_A] Stoichiometric air-to-fuel mass ratio 
        self.theta        = 0.074    # [sec**-1]         Contact time
        self.LHV_F        = 43.435   # [kJ/g-mol]        Lower heating value of Jet-A
        self.LHV_H2       = 240.2    # [kJ/g-mol]        Lower heating value of Hydrogen
        self.LHV_CO       = 283.1    # [kJ/g-mol]        Lower heating value of Carbon Monoxide
        self.V_cat        = 9.653    # [cm**3]           Catalyst bed volume
        self.eta          = 0.9


# RCAIDE/Library/Missions/Segments/Untrimmed/Untrimmed.py
# 
# 
# Created:  Jul 2023, M. Clarke 
 
# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------  

import numpy as np

# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------   
def initialize_conditions(segment):
    """
    Initializes conditions for fixed speed and altitude trim analysis using AVL

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed
            - altitude : float
                Flight altitude [m]
            - air_speed : float
                True airspeed [m/s]
            - sideslip_angle : float
                Aircraft sideslip angle [rad]
            - linear_acceleration_x : float
                Acceleration in x-direction [m/s^2]
            - linear_acceleration_y : float
                Acceleration in y-direction [m/s^2]
            - linear_acceleration_z : float
                Acceleration in z-direction [m/s^2]
            - roll_rate : float
                Aircraft roll rate [rad/s]
            - pitch_rate : float
                Aircraft pitch rate [rad/s]
            - yaw_rate : float
                Aircraft yaw rate [rad/s]
            - state:
                conditions : Data
                    State conditions container
                initials : Data, optional
                    Initial conditions from previous segment
    
    Returns
    -------
    None
        Updates segment conditions directly:
            - conditions.freestream.altitude [m]
            - conditions.frames.inertial.position_vector [m]
            - conditions.frames.inertial.velocity_vector [m/s]
            - conditions.frames.inertial.acceleration_vector [m/s^2]
            - conditions.static_stability.roll_rate [rad/s]
            - conditions.static_stability.pitch_rate [rad/s]
            - conditions.static_stability.yaw_rate [rad/s]
    
    Notes
    -----
    This function sets up the initial conditions for a single point analysis with
    fixed speed and altitude. It includes trim state variables for AVL (Athena 
    Vortex Lattice) aerodynamic analysis.

    **Calculation Process**
        1. Check initial conditions
        2. Decompose velocity into components using sideslip angle:
                - v_x = V * cos(β)
                - v_y = V * sin(β)
            where:
                - V is true airspeed
                - β is sideslip angle
        3. Set position and altitude
        4. Set acceleration vector
        5. Set angular rates for trim analysis

    **Major Assumptions**
        * Steady state trim condition
        * Small angle approximations
        * Linear aerodynamics
        * No atmospheric variations

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    RCAIDE.Methods.Aerodynamics.AVL
    """      
    
    # unpack
    alt                    = segment.altitude
    air_speed              = segment.air_speed  
    sideslip               = segment.sideslip_angle
    linear_acceleration_x  = segment.linear_acceleration_x
    linear_acceleration_y  = segment.linear_acceleration_y 
    linear_acceleration_z  = segment.linear_acceleration_z
    roll_rate              = segment.roll_rate
    pitch_rate             = segment.pitch_rate
    yaw_rate               = segment.yaw_rate
    
    
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
    
    # pack
    air_speed_x                                                   = np.cos(sideslip)*air_speed 
    air_speed_y                                                   = np.sin(sideslip)*air_speed 
    segment.state.conditions.freestream.altitude[:,0]             = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down
    segment.state.conditions.frames.inertial.velocity_vector[:,0] = air_speed_x
    segment.state.conditions.frames.inertial.velocity_vector[:,1] = air_speed_y
    segment.state.conditions.frames.inertial.acceleration_vector  = np.array([[linear_acceleration_x,linear_acceleration_y,linear_acceleration_z]])  
    segment.state.conditions.static_stability.roll_rate[:,0]      = roll_rate         
    segment.state.conditions.static_stability.pitch_rate[:,0]     = pitch_rate
    segment.state.conditions.static_stability.yaw_rate[:,0]       = yaw_rate      


# RCAIDE/Library/Missions/Segments/Hover/Climb.py
# 
# 
# Created:  Jul 2023, M. Clarke 
 
# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------   
def initialize_conditions(segment):
    """
    Initializes conditions for vertical climb segment

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed
            - altitude_start : float
                Initial altitude [m]
            - altitude_end : float
                Final altitude [m]
            - climb_rate : float
                Vertical climb rate [m/s]
            - state:
                numerics:
                    dimensionless:
                        control_points : array
                            Discretization points [-]
                conditions:
                    frames:
                        inertial:
                            time : array
                                Time points [s]
                initials : Data, optional
                    Initial conditions from previous segment

    Returns
    -------
    None
        Updates segment conditions directly:
            - conditions.frames.inertial.velocity_vector [m/s]
            - conditions.frames.inertial.position_vector [m]
            - conditions.freestream.altitude [m]
            - conditions.frames.inertial.time [s]

    Notes
    -----
    This function sets up the initial conditions for a vertical climb segment with
    constant climb rate. The segment handles pure vertical motion with no horizontal
    velocity components.

    **Calculation Process**
        1. Check initial conditions
        2. Discretize altitude profile:
            alt = alt0 + (altf - alt0)*t_norm
        3. Calculate time required:
            dt = (altf - alt0)/climb_rate
        4. Set velocity components:
            - v_x = 0
            - v_z = -climb_rate (z points down)
        5. Scale time points:
        t = t_norm * dt

    **Major Assumptions**
        * Constant climb rate
        * Pure vertical motion
        * No horizontal velocity
        * No atmospheric variations
        * Quasi-steady climb

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    """        
    
    # unpack
    climb_rate = segment.climb_rate
    alt0       = segment.altitude_start 
    altf       = segment.altitude_end
    t_nondim   = segment.state.numerics.dimensionless.control_points
    t_initial  = segment.state.conditions.frames.inertial.time[0,0]
    conditions = segment.state.conditions  

    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]

    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0
    
    # process velocity vector
    v_z   = -climb_rate # z points down    
    dt = (altf - alt0)/climb_rate

    # rescale operators
    t = t_nondim * dt

    # pack
    t_initial = segment.state.conditions.frames.inertial.time[0,0]
    segment.state.conditions.frames.inertial.time[:,0] = t_initial + t[:,0]    
    
    # pack conditions    
    conditions.frames.inertial.velocity_vector[:,0]         = 0.
    conditions.frames.inertial.velocity_vector[:,2]         = v_z
    conditions.frames.inertial.position_vector[:,2]         = -alt[:,0] # z points down
    conditions.freestream.altitude[:,0]                     =  alt[:,0] # positive altitude in this context
    segment.state.conditions.frames.inertial.time[:,0]      = t_initial + t[:,0]


# RCAIDE/Library/Methods/Performance/compute_noise_certification_data.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Core import Units , Data   
import matplotlib.colors
import matplotlib.colors as colors
from RCAIDE.Library.Plots import *
 
# Pacakge imports 
import numpy as np
from matplotlib import pyplot as plt
import plotly.graph_objects as go
 
# ----------------------------------------------------------------------
#  Compute Aircraft Noise Certification Data  
# ----------------------------------------------------------------------  
def compute_noise_certification_data(approach_mission  = None,
                                     takeoff_mission   = None,
                                     noise_level       = None,
                                     min_noise_level   = 45,  
                                     max_noise_level   = 105, 
                                     noise_scale_label = "Max. SPL [dbA]",
                                     save_figure       = False,
                                     show_figure       = True,
                                     save_filename     = "Certification_Noise", 
                                     colormap          = 'jet',
                                     file_type         = ".png",
                                     width             = 12, 
                                     height            = 6):
    """Calculates the noise at certification points as well as the noise contours of approach and takeoff.
    A combined approach-takeoff noisec contour is also created 
    """ 
            
    if approach_mission == None:
        raise AssertionError('Approach mission not specifed!')
    if takeoff_mission == None:
        raise AssertionError('Takeoff mission not specifed!')
     
    microphone_x_resolution                = 401 
    microphone_y_resolution                = 9  
    noise_times_steps                      = 51 
    number_of_microphone_in_stencil        = 1800
    
    # update weights analysis
    for segment in approach_mission.segments:
        if segment.analyses.noise == None:
            raise AssertionError('Noise analysis not specifed!')
        noise_analysis = segment.analyses.noise
        noise_analysis.settings.microphone_x_resolution                = microphone_x_resolution
        noise_analysis.settings.microphone_y_resolution                = microphone_y_resolution
        noise_analysis.settings.noise_times_steps                      = noise_times_steps
        noise_analysis.settings.number_of_microphone_in_stencil        = number_of_microphone_in_stencil
        noise_analysis.settings.microphone_min_y                       = 1E-6   
        noise_analysis.settings.microphone_max_y                       = 1800
        noise_analysis.settings.microphone_min_x                       = 1E-6   
        noise_analysis.settings.microphone_max_x                       = 8000
    
    # update weights analysis
    for segment in takeoff_mission.segments:
        if segment.analyses.noise == None:
            raise AssertionError('Noise analysis not specifed!')
        noise_analysis = segment.analyses.noise 
        noise_analysis.settings.microphone_x_resolution                = microphone_x_resolution
        noise_analysis.settings.microphone_y_resolution                = microphone_y_resolution
        noise_analysis.settings.noise_times_steps                      = noise_times_steps
        noise_analysis.settings.number_of_microphone_in_stencil        = number_of_microphone_in_stencil
        noise_analysis.settings.microphone_min_y                       = 1E-6   
        noise_analysis.settings.microphone_max_y                       = 1800
        noise_analysis.settings.microphone_min_x                       = -2000 + 1E-6  
        noise_analysis.settings.microphone_max_x                       = 6000
    
    # evaluate both missions
    approach_results = approach_mission.evaluate() 
    takeoff_results  = takeoff_mission.evaluate()
    
    # post process results
    noise_data  =  post_process_certification_noise_data(approach_results,takeoff_results)
    
    # plot diagram
    if show_figure:  
        fig = plt.figure(save_filename)
        fig.set_size_inches(width,height) 
        
        noise_levels   = np.linspace(min_noise_level,max_noise_level,7)  
        noise_cmap     = plt.get_cmap('turbo')
        noise_new_cmap = truncate_colormap(noise_cmap,0.0, 1.0) 

        noise_level = noise_data.certification_SPL_dBA_max           
        X           = noise_data.certification_microphone_locations[:,:,0]
        Y           = noise_data.certification_microphone_locations[:,:,1] 
        
        ap_noise_level = noise_data.approach_SPL_dBA_max           
        ap_X           = noise_data.approach_microphone_locations[:,:,0]
        ap_Y           = noise_data.approach_microphone_locations[:,:,1]
        ap_POS         = noise_data.approach_trajectory

        to_noise_level = noise_data.takeoff_SPL_dBA_max           
        to_X           = noise_data.takeoff_microphone_locations[:,:,0]
        to_Y           = noise_data.takeoff_microphone_locations[:,:,1]
        to_POS         = noise_data.takeoff_trajectory
        
        axis_0   = fig.add_subplot(2,2,1) 
        axis_0.set_xlabel('x [m]')
        axis_0.set_ylabel('altitude [m]')   
        axis_1   = fig.add_subplot(2,2,3) 
        axis_1.set_xlabel('x [m]')
        axis_1.set_ylabel('y [m]')   
        axis_2   = fig.add_subplot(2,2,2) 
        axis_2.set_xlabel('x [m]')
        axis_2.set_ylabel('y [m]')   
        axis_3   = fig.add_subplot(2,2,4) 
        axis_3.set_xlabel('x [m]')
        axis_3.set_ylabel('y [m]')
        
        # plot aircraft position
        axis_0.plot(ap_POS[:,0],-ap_POS[:,2], color = 'black', linestyle = '-' , marker = 'o', linewidth = 2, label= "Approach")
        axis_0.plot(to_POS[:,0],-to_POS[:,2], color = 'blue', linestyle = '-' , marker = 's', linewidth = 2, label= "Takeoff")
        axis_0.legend(loc='upper center') 
    
        # plot aircraft noise levels   
        CS_11    = axis_1.contourf(X,Y,noise_level ,noise_levels,cmap = noise_new_cmap,extend='both')  
        CS_12    = axis_1.contourf(X,-Y,noise_level ,noise_levels,cmap = noise_new_cmap,extend='both')
        CS_21    = axis_2.contourf(ap_X,ap_Y,ap_noise_level ,noise_levels,cmap = noise_new_cmap,extend='both')  
        CS_22    = axis_2.contourf(ap_X,-ap_Y,ap_noise_level ,noise_levels,cmap = noise_new_cmap,extend='both')
        CS_31    = axis_3.contourf(to_X,to_Y,to_noise_level ,noise_levels,cmap = noise_new_cmap,extend='both')  
        CS_32    = axis_3.contourf(to_X,-to_Y,to_noise_level ,noise_levels,cmap = noise_new_cmap,extend='both')
         
        cbar  = fig.colorbar(CS_11, ax=axis_1)        
        cbar.ax.set_ylabel(noise_scale_label, rotation =  90)

        cbar  = fig.colorbar(CS_11, ax=axis_2)        
        cbar.ax.set_ylabel(noise_scale_label, rotation =  90)

        cbar  = fig.colorbar(CS_11, ax=axis_3)        
        cbar.ax.set_ylabel(noise_scale_label, rotation =  90)     

        run_way_x_pts = np.linspace(0, 3000, 20)
        run_way_y_pts = run_way_x_pts * 0

        axis_1.plot(run_way_x_pts,run_way_y_pts, color = 'grey', linestyle = '-' , linewidth = 5 , alpha=0.5)
        axis_2.plot(run_way_x_pts,run_way_y_pts, color = 'grey', linestyle = '-' , linewidth = 5 , alpha=0.5)
        axis_3.plot(run_way_x_pts,run_way_y_pts, color = 'grey', linestyle = '-' , linewidth = 5 , alpha=0.5)
        
        set_axes(axis_1)
        set_axes(axis_2)
        set_axes(axis_3)

        axis_0.set_title('Flight Trajectory')        
        axis_1.set_title('Approach and Takeoff Noise')
        axis_2.set_title('Approach Noise')
        axis_3.set_title('Takeoff Noise')
        
        fig.tight_layout()  
        if save_figure: 
            figure_title  = save_filename
            plt.savefig(figure_title + file_type )   
        
    return noise_data 
    
def post_process_certification_noise_data(approach_results,takeoff_results): 
   
    approach_noise_data   = post_process_noise_data(approach_results)
    takeoff_noise_data    = post_process_noise_data(takeoff_results) 
    
    # append approach noise                                
    approach_pos         = approach_noise_data.aircraft_position
    approach_pos[:,0]   -= 2000 
    
    # append takeoff noise  
    cert_SPL_dBA_max  = np.max(np.concatenate((approach_noise_data.SPL_dBA,takeoff_noise_data.SPL_dBA), axis = 0) ,axis = 0)                      
    cert_pos          = np.concatenate((approach_pos, takeoff_noise_data.aircraft_position), axis = 0)     
    cert_mic_locs     = takeoff_noise_data.microphone_locations 
     
    res = Data(
        certification_SPL_dBA_max         = cert_SPL_dBA_max, 
        certification_trajectory          = cert_pos,     
        certification_microphone_locations= cert_mic_locs,

        approach_SPL_dBA_max              = np.max(approach_noise_data.SPL_dBA,axis = 0),
        approach_SPL_dBA                  = approach_noise_data.SPL_dBA, 
        approach_time                     = approach_noise_data.time,   
        approach_trajectory               = approach_pos,     
        approach_microphone_locations     = takeoff_noise_data.microphone_locations, 

        takeoff_SPL_dBA_max               = np.max(takeoff_noise_data.SPL_dBA,axis = 0),
        takeoff_SPL_dBA                   = takeoff_noise_data.SPL_dBA, 
        takeoff_time                      = takeoff_noise_data.time,   
        takeoff_trajectory                = takeoff_noise_data.aircraft_position,     
        takeoff_microphone_locations      = takeoff_noise_data.microphone_locations     
    
    )
    
    print('Certification Noise')
    print('-----------------------------------------------')
    print('2000 m  Approach Noise   :', round(cert_SPL_dBA_max[0, 0], 2)) 
    print('6000 m  Flyover Noise    :', round(cert_SPL_dBA_max[-1, 0], 2))
    print('450  m  Sideline Noise   :', round(max(cert_SPL_dBA_max[:, 2]), 2))
    
    area_85_dbA =  len(np.where(cert_SPL_dBA_max.flatten()>85)[0]) *100 / len(cert_SPL_dBA_max.flatten()) 
    area_65_dbA =  len(np.where(cert_SPL_dBA_max.flatten()>65)[0]) *100 / len(cert_SPL_dBA_max.flatten())
    print('% Area > 85 dBA Threshold:', round(area_85_dbA, 2)) 
    print('% Area > 65 dBA Threshold:', round(area_65_dbA, 2))
    
    res.approach_noise_2000m = cert_SPL_dBA_max[0, 0] 
    res.flyover_noise_6000m  = cert_SPL_dBA_max[-1, 0] 
    res.sideline_noise_450m  = max(cert_SPL_dBA_max[:, 2])
    res.area_65_dbA = area_65_dbA
    res.area_85_dbA = area_85_dbA 
    
    return res


# ------------------------------------------------------------------ 
# Truncate colormaps
# ------------------------------------------------------------------  
def truncate_colormap(cmap, minval=0.0, maxval=1.0, n=100):
    new_cmap = colors.LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmap.name, a=minval, b=maxval),
        cmap(np.linspace(minval, maxval, n)))
    return new_cmap



# RCAIDE/Library/Attributes/Coolants/Glycol_Water.py
# 
#
# Created:  Mar 2024, M. Clarke

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 
from .Coolant import Coolant

# ---------------------------------------------------------------------------------------------------------------------- 
#  Glycol_Water
# ----------------------------------------------------------------------------------------------------------------------  
class Glycol_Water(Coolant):
    """
    Implementation of a 50-50 ethylene glycol-water mixture coolant properties.

    This class provides thermophysical properties and calculation methods for a standard
    ethylene glycol-water mixture commonly used in cooling systems.

    Attributes
    ----------
    tag : str
        Identifier set to 'Glycol_Water'   
    percent_glycol : float
        Mass fraction of glycol in the mixture (default: 0.5)
    density : float
        Mass per unit volume [kg/m³]
    specific_heat_capacity : float
        Heat capacity at constant pressure [J/kg·K]
    thermal_conductivity : float
        Heat conduction coefficient [W/m·K]
    dynamic_viscosity : float
        Absolute viscosity [Pa·s]
    Prandtl_number : float
        Dimensionless number for heat transfer characteristics
    kinematic_viscosity : float
        Ratio of dynamic viscosity to density [m²/s]

    Notes
    -----
    All properties are currently implemented as constant values, though the method
    structure allows for future implementation of temperature and pressure dependence.

    **Major Assumptions**
        * Mixture is exactly 50% water and 50% ethylene glycol by mass
        * Standard atmospheric pressure conditions
        * Single-phase liquid mixture

    **Extra modules required**
    None

    References
    ----------
    [1] The Engineering ToolBox (2003). Ethylene Glycol Heat-Transfer Fluid Properties. [online] Available at: https://www.engineeringtoolbox.com/ethylene-glycol-d_146.html [Accessed 8 January 2025].
    [2] Microelectronics Heat Transfer Laboratory. (1997). Fluid Properties Calculator. Fluid properties calculator. http://www.mhtl.uwaterloo.ca/old/onlinetools/airprop/airprop.html 

    See Also
    --------
    RCAIDE.Library.Components.Thermal_Management : Thermal management system components
    """

    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            Mixture is 50% water-50% ethylene-glycol
        
        Source: 
            Engineering Toolbox: https://www.engineeringtoolbox.com/ethylene-glycol-d_146.html
            University of Waterloo: http://www.mhtl.uwaterloo.ca/old/onlinetools/airprop/airprop.html 
        """ 
        self.tag                       = 'Glycol_Water'
        self.percent_glycol            = 0.5 
        self.density                   = 1075                       # kg/m^3
        self.specific_heat_capacity    = 3300                       # J/kg.K
        self.thermal_conductivity      = 0.387                      # W/m.K
        self.dynamic_viscosity         = 0.0019                     # Pa.s
        self.Prandtl_number            = self.specific_heat_capacity * self.dynamic_viscosity / self.thermal_conductivity
        self.kinematic_viscosity       = self.dynamic_viscosity / self.density

    def compute_cp(self,T=300):
        """         
        Calculate specific heat capacity of glycol-water mixture at given temperature.

        Parameters
        ----------
        T : float, optional
            Temperature [K] (default: 300)

        Returns
        -------
        cp : float
            Specific heat capacity [J/(kg·K)]

        Notes
        -----
        Currently returns constant value of 3300 J/(kg·K), based on the default value, regardless of temperature.

        **Major Assumptions**
            * Temperature independence
            * 50-50 water-glycol mixture

        References
        ----------
        [1] The Engineering ToolBox (2003). Ethylene Glycol Heat-Transfer Fluid Properties. [online] Available at: https://www.engineeringtoolbox.com/ethylene-glycol-d_146.html [Accessed 8 January 2025].
  
        """ 
        return self.specific_heat_capacity
    
    def compute_absolute_viscosity(self,T=300.,p=101325.):
        """ 
        Calculate dynamic viscosity of glycol-water mixture at given conditions.

        Parameters
        ----------
        T : float, optional
            Temperature [K] (default: 300)
        p : float, optional
            Pressure [Pa] (default: 101325)

        Returns
        -------
        mu : float
            Dynamic viscosity [kg/(m·s)]

        Notes
        -----
        Currently returns constant value of 0.0019 Pa·s, based on the default value, regardless of conditions.

        **Major Assumptions**
            * Temperature and pressure independence
            * 50-50 water-glycol mixture

        References
        ----------
        [1] The Engineering ToolBox (2003). Ethylene Glycol Heat-Transfer Fluid Properties. [online] Available at: https://www.engineeringtoolbox.com/ethylene-glycol-d_146.html [Accessed 8 January 2025].

        """ 
        return  self.dynamic_viscosity
    
    def compute_density(self,T=300.,p=101325.):  
        """ 
        Calculate density of glycol-water mixture at given conditions.

        Parameters
        ----------
        T : float, optional
            Temperature [K] (default: 300)
        p : float, optional
            Pressure [Pa] (default: 101325)

        Returns
        -------
        rho : float
            Density [kg/m³]

        Notes
        -----
        Currently returns constant value of 1075 kg/m³ based on the default value.

        **Major Assumptions**
            * Temperature and pressure independence
            * 50-50 water-glycol mixture

        References
        ----------
        [1] The Engineering ToolBox (2003). Ethylene Glycol Heat-Transfer Fluid Properties. [online] Available at: https://www.engineeringtoolbox.com/ethylene-glycol-d_146.html [Accessed 8 January 2025].
   
        """         
        return self.density  
    
    def compute_thermal_conductivity(self,T=300.,p=101325.): 
        """
        Calculate thermal conductivity of glycol-water mixture at given conditions.

        Parameters
        ----------
        T : float, optional
            Temperature [K] (default: 300)
        p : float, optional
            Pressure [Pa] (default: 101325)

        Returns
        -------
        k : float
            Thermal conductivity [W/(m·K)]

        Notes
        -----
        Currently returns constant value of 0.387 W/(m·K), based on the default values, 
        regardless of conditions.

        **Major Assumptions**
            * Temperature and pressure independence
            * 50-50 water-glycol mixture

        References
        ----------
        [1] Microelectronics Heat Transfer Laboratory. (1997). Fluid Properties Calculator. Fluid properties calculator. http://www.mhtl.uwaterloo.ca/old/onlinetools/airprop/airprop.html 
       
        """          
        return self.thermal_conductivity
    
    
    def compute_prandtl_number(self,T=300.): 
        """
        Calculate Prandtl number of glycol-water mixture at given temperature.

        Parameters
        ----------
        T : float, optional
            Temperature [K] (default: 300)

        Returns
        -------
        Pr : float
            Prandtl number [dimensionless]

        Notes
        -----
        Calculated using the relationship:
        Pr = (μ·Cp)/k

        where:
            - μ is dynamic viscosity
            - Cp is specific heat capacity
            - k is thermal conductivity

        **Major Assumptions**
            * Temperature independence of constituent properties
            * 50-50 water-glycol mixture

        References
        ----------
        [1] Microelectronics Heat Transfer Laboratory. (1997). Fluid Properties Calculator. Fluid properties calculator. http://www.mhtl.uwaterloo.ca/old/onlinetools/airprop/airprop.html 
       
        """          
        
        Cp = self.compute_cp(T)
        mu = self.compute_absolute_viscosity(T)
        K  = self.compute_thermal_conductivity(T)
        return  mu*Cp/K      


# RCAIDE/Library/Methods/Powertrain/Converters/Expansion_Nozzle/__init__.py

"""
This module provides functionality for modeling expansion nozzles in powertrains. It includes methods for computing 
expansion nozzle performance and appending expansion nozzle conditions to simulation results.

See Also
--------
RCAIDE.Library.Methods.Powertrain.Converters
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .append_expansion_nozzle_conditions    import append_expansion_nozzle_conditions         
from .compute_expansion_nozzle_performance  import compute_expansion_nozzle_performance


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Raymer/compute_horizontal_tail_weight.py
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import RCAIDE 
from RCAIDE.Framework.Core    import Units

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Horizontal Tail Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_horizontal_tail_weight(vehicle, wing, settings,elevator_fraction=0.4):
    """
    Calculates horizontal tail weight based on Raymer's empirical method.

    Parameters
    ----------
    vehicle : RCAIDE.Vehicle()
        Vehicle data structure containing:
            - mass_properties.max_takeoff : float
                Maximum takeoff weight [kg]
            - flight_envelope.ultimate_load : float
                Ultimate load factor
            - wings['main_wing'] : Data()
                Main wing properties including origin and aerodynamic center
            - fuselages['fuselage'].width : float
                Width of the fuselage [m]
    wing : RCAIDE.Component()
        Horizontal tail component containing:
            - areas.reference : float
                Tail surface area [m^2]
            - origin : array
                Location of tail measured from nose [m]
            - aerodynamic_center : array
                Location of ac measured from leading edge [m]
            - sweeps.quarter_chord : float
                Quarter chord sweep angle [rad]
            - thickness_to_chord : float
                Thickness-to-chord ratio
            - spans.projected : float
                Projected span [m]
            - aspect_ratio : float
                Aspect ratio
    elevator_fraction : float, optional
        Fraction of horizontal tail area that is elevator, defaults to 0.4
    settings.advanced_composites : bool, optional
        True if advanced composites are used, False otherwise

    Returns
    -------
    tail_weight : float
        Weight of the horizontal tail [kg]

    Notes
    -----
    This method implements Raymer's correlation for horizontal tail weight estimation,
    accounting for geometry, loads, and configuration effects.

    **Major Assumptions**
        * Not an all-moving horizontal tail (Kuht = 1.0)
        * Elevator comprises 40% of tail area by default
        * Correlation based on transport category aircraft data
        * If advanced composites are used, the tail weight is reduced by 15%

    **Theory**
    The horizontal tail weight is calculated using:
    .. math::
        W_{ht} = 0.0379K_{uht}(1 + \frac{F_w}{B_h})^{-0.25}W_{dg}^{0.639}N_{ult}^{0.1}S_{ht}^{0.75}L_t^{-1}K_y^{0.704}\cos(\Lambda)^{-1}AR^{0.166}(1 + \frac{S_e}{S_{ht}})^{0.1}

    where:
        * :math:`K_{uht}` is 1.0 for fixed tails, 1.143 for all-moving
        * :math:`F_w` is fuselage width
        * :math:`B_h` is horizontal tail span
        * :math:`L_t` is tail arm length
        * :math:`K_y` is a function of tail arm length
        * :math:`\Lambda` is quarter-chord sweep angle

    References
    ----------
    [1] Raymer, D., "Aircraft Design: A Conceptual Approach", AIAA 
        Education Series, 2018. 

    See Also
    --------
    RCAIDE.Library.Methods.Weights.Correlation_Buildups.Raymer.compute_vertical_tail_weight
    RCAIDE.Library.Methods.Weights.Correlation_Buildups.Raymer.compute_operating_empty_weight
    """

    ref_wing = None 
    for wing in  vehicle.wings:
        if isinstance(wing, RCAIDE.Library.Components.Wings.Main_Wing):
            ref_wing  =  wing
    
    S = 0
    if ref_wing == None:
        for wing in  vehicle.wings:
            if S < wing.areas.reference:
                ref_wing = wing
    
    L_fuselage = 0
    for fuselage in vehicle.fuselages:
        if L_fuselage < fuselage.lengths.total:
            ref_fuselage = fuselage
            L_fuselage = ref_fuselage.lengths.total
    diff =[]
    names = []
    horiz_tail_location = vehicle.wings.horizontal_stabilizer.origin[0][0]
    for segment in ref_fuselage.segments:
        segment_loc = segment.percent_x_location * ref_fuselage.lengths.total
        diff.append(np.abs(segment_loc - horiz_tail_location))
        names.append(segment.tag)
    min_value = min(diff)
    min_index = diff.index(min_value)
            
    Kuht    = 1 # not an all-moving unit horizontal tail
    Fw      = ref_fuselage.segments[names[min_index]].width / Units.ft
    Bh      = wing.spans.projected / Units.ft
    DG      = vehicle.mass_properties.max_takeoff / Units.lbs
    Sht     = wing.areas.reference / Units.ft ** 2
    Lt      = (wing.origin[0][0] + wing.aerodynamic_center[0] - ref_wing.origin[0][0] -
                ref_wing.aerodynamic_center[0]) / Units.ft
    Ky      = 0.3 * Lt
    sweep   = wing.sweeps.quarter_chord
    Ah      = wing.aspect_ratio
    Se      = elevator_fraction * Sht

    tail_weight = 0.0379 * Kuht * (1 + Fw / Bh) ** (-0.25) * DG ** 0.639 *\
                  vehicle.flight_envelope.ultimate_load ** 0.1 * Sht ** 0.75 * Lt ** -1 *\
                  Ky ** 0.704 * np.cos(sweep) ** (-1) * Ah ** 0.166 * (1 + Se / Sht) ** 0.1
    
    if settings.advanced_composites:
        tail_weight *= 0.85

    return tail_weight * Units.lbs


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/compute_fuselage_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import RCAIDE 
from RCAIDE.Framework.Core    import Units

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
# Fuselage Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_fuselage_weight(vehicle):
    """
    Computes the fuselage weight using NASA FLOPS weight estimation method. Accounts for 
    aircraft type.

    Parameters
    ----------
    vehicle : Vehicle
        The vehicle instance containing:
            - networks : list
                Propulsion system data for engine count and mounting
            - fuselages['fuselage'] : Fuselage
                Primary fuselage with:
                    - lengths.total : float
                        Total fuselage length [m]
                    - width : float
                        Maximum fuselage width [m]
                    - heights.maximum : float
                        Maximum fuselage height [m]
            - flight_envelope.ultimate_load : float
                Ultimate load factor (default: 3.75)
            - systems.accessories : str
                Aircraft type ('short-range', 'commuter', 'medium-range', 
                'long-range', 'sst', 'cargo')
            - mass_properties.max_takeoff : float
                Maximum takeoff weight [kg]
            - design_mach_number : float
                Design cruise Mach number

    Returns
    -------
    WFUSE : float
        Fuselage structural weight [kg]

    Notes
    -----
    Uses FLOPS correlations developed from transport aircraft database. For more details, 
    please refer to the FLOPS documentation: https://ntrs.nasa.gov/citations/20170005851  

    **Major Assumptions**
        * Single fuselage configuration (NFUSE = 1). If there are multiple fuselages, these are calcualted separately.
        * Fuselage is tagged as 'fuselage'
    
    **Theory**
    For transport aircraft:
    .. math::
         W_{fuse} = 1.35(L_{f}D_{eq})^{1.28}(1 + 0.05*N_{ef})(1 + 0.38*F_{cargo})
 
    where:
        * L_f = fuselage length
        * D_eq = equivalent diameter
        * N_ef = number of fuselage-mounted engines
        * F_cargo = cargo aircraft flag

    References
    ----------
    [1] NASA Flight Optimization System (FLOPS)

    See Also
    --------
    RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.Transport.FLOPS.compute_operating_empty_weight
    """
    
    L =  0
    for fuselage in vehicle.fuselages:
        if L < fuselage.lengths.total: 
            total_length = fuselage.lengths.total
            width        = fuselage.width
            max_height   = fuselage.heights.maximum
    
    XL  = total_length / Units.ft  # Fuselage length, ft
    DAV = (width + max_height) / 2. * 1 / Units.ft
    NENG = 0
    FNEW = 0
    FNEF = 0 
    for network in  vehicle.networks:
        for propulsor in network.propulsors:
            if isinstance(propulsor, RCAIDE.Library.Components.Powertrain.Propulsors.Turbofan) or  isinstance(propulsor, RCAIDE.Library.Components.Powertrain.Propulsors.Turbojet):
                NENG += 1 
                if propulsor.wing_mounted: 
                    FNEW += 1  
                else:
                    FNEF += 1
    if vehicle.systems.accessories == 'freighter':
        CARGF = 1
    else:
        CARGF = 0  # Cargo aircraft floor factor [0 for passenger transport, 1 for cargo transport
    NFUSE = 1  # Number of fuselages
    WFUSE = 1.35 * (XL * DAV) ** 1.28 * (1 + 0.05 * FNEF) * (1 + 0.38 * CARGF) * NFUSE
    
    return WFUSE * Units.lbs


# RCAIDE/Framework/Analyses/Mission/Segments/Climb/Constant_Speed_Constant_Rate.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
from RCAIDE.Framework.Core                            import Units 
from RCAIDE.Framework.Mission.Segments.Evaluate       import Evaluate
from RCAIDE.Library.Mission                           import Common,Segments

# ----------------------------------------------------------------------------------------------------------------------
#  Constant_Speed_Constant_Rate
# ---------------------------------------------------------------------------------------------------------------------- 
class Constant_Speed_Constant_Rate(Evaluate):
    """ The most basic segment. Fly at a constant true airspeed at a fixed rate of climb between 2 altitudes.
    
        Assumptions:
        None
        
        Source:
        None
    """       
    
    def __defaults__(self):
        """ This sets the default solver flow. Anything in here can be modified after initializing a segment.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """          
        
        # -------------------------------------------------------------------------------------------------------------- 
        #   User Inputs
        # -------------------------------------------------------------------------------------------------------------- 
        self.altitude_start     = None # Optional
        self.altitude_end       = 10. * Units.km
        self.climb_rate         = 3.  * Units.m / Units.s
        self.air_speed          = 100 * Units.m / Units.s
        self.true_course        = 0.0 * Units.degrees     
        
        # -------------------------------------------------------------------------------------------------------------- 
        #  Mission specific processes 
        # --------------------------------------------------------------------------------------------------------------   
        initialize                         = self.process.initialize  
        initialize.differentials_altitude  = Common.Initialize.differentials_altitude
        initialize.conditions              = Segments.Climb.Constant_Speed_Constant_Rate.initialize_conditions
        iterate                            = self.process.iterate
        iterate.unknowns.controls          = Common.Unpack_Unknowns.control_surfaces
        iterate.residuals.flight_dynamics  = Common.Residuals.flight_dynamics
        iterate.unknowns.mission           = Common.Unpack_Unknowns.orientation   
        
        return
       


# RCAIDE/Library/Missions/Segments/Descent/Constant_Speed_Constant_Rate.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------  
#  IMPORT 
# ----------------------------------------------------------------------------------------------------------------------  
# package imports 
import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------  
def initialize_conditions(segment):
    """
    Initializes conditions for constant speed descent at fixed rate

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed
            - descent_rate : float
                Rate of descent [m/s]
            - air_speed : float
                True airspeed to maintain [m/s]
            - altitude_start : float
                Initial altitude [m]
            - altitude_end : float
                Final altitude [m]
            - sideslip_angle : float
                Aircraft sideslip angle [rad]
            - state:
                numerics.dimensionless.control_points : array
                    Discretization points [-]
                conditions : Data
                    State conditions container
    
    Returns
    -------
    None
    
    Notes
    -----
    This function sets up the initial conditions for a descent segment with constant
    true airspeed and constant descent rate. The horizontal velocity components are
    determined from the airspeed and descent rate constraints. Updates segment conditions
    directly with velocity_vector [m/s], altitude [m], and position_vector [m].

    **Calculation Process**
        1. Discretize altitude profile
        2. Calculate horizontal velocity magnitude:
        v_xy = sqrt(V^2 - v_z^2) where:
            - V is true airspeed
            - v_z is descent rate
        3. Decompose horizontal velocity using sideslip angle:
            - v_x = v_xy * cos(β)
            - v_y = v_xy * sin(β)

            where β is sideslip angle

    **Major Assumptions**
        * Constant true airspeed
        * Constant descent rate
        * Small angle approximations
        * Quasi-steady flight
        * No wind effects

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    """     
    
    # unpack
    descent_rate = segment.descent_rate
    air_speed    = segment.air_speed   
    alt0         = segment.altitude_start 
    altf         = segment.altitude_end
    beta         = segment.sideslip_angle
    t_nondim     = segment.state.numerics.dimensionless.control_points
    conditions   = segment.state.conditions  

    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2] 
        
    # check for initial velocity vector
    if air_speed is None:
        if not segment.state.initials: raise AttributeError('initial airspeed not set')
        air_speed  =  np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1,:])   
            
    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0
    
    # process velocity vector
    v_xy_mag    = air_speed
    v_z         = descent_rate # z points down
    v_xy        = np.sqrt(v_xy_mag**2 - v_z**2 )
    v_x         = np.cos(beta)*v_xy
    v_y         = np.sin(beta)*v_xy
    
    # pack conditions    
    conditions.frames.inertial.velocity_vector[:,0] = v_x
    conditions.frames.inertial.velocity_vector[:,1] = v_y
    conditions.frames.inertial.velocity_vector[:,2] = v_z
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0] # z points down
    conditions.freestream.altitude[:,0]             =  alt[:,0] # positive altitude in this context


# RCAIDE/Library/Missions/Common/Pre_Process/aerodynamics.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from RCAIDE.Library.Methods.Geometry.Planform  import wing_segmented_planform, wing_planform

# ----------------------------------------------------------------------------------------------------------------------
#  aerodynamics
# ----------------------------------------------------------------------------------------------------------------------  
def aerodynamics(mission):
    """
    Initializes and processes aerodynamic models for mission segments

    Parameters
    ----------
    mission : Mission
        The mission containing segments to be analyzed
            - analyses.aerodynamics : Analysis
                Aerodynamic analysis module
                - vehicle : Vehicle
                    Aircraft geometry definition
                    - wings : list
                        Wing geometry definitions
                - process.compute.lift.inviscid_wings : Process
                    Lift computation process
                - surrogates : Data
                    Aerodynamic surrogate models
                - reference_values : Data
                    Reference aerodynamic parameters

    Notes
    -----
    This function prepares the aerodynamic analysis for each mission segment.
    It ensures proper wing geometry computation and manages aerodynamic
    surrogate models across segments for computational efficiency.

    The function performs the following steps:
        1. Computes wing planform properties
        2. Reuses previous segment's aerodynamic data when possible
        3. Initializes new aerodynamic analyses when needed

    **Wing Processing**
    
    For each wing:
        - If multi-segmented: Uses wing_segmented_planform
        - If single segment: Uses wing_planform

    **Major Assumptions**
        * Valid wing geometry definitions
        * Compatible aerodynamic models between segments
        * Proper initialization of first segment
        * Continuous aerodynamic characteristics

    Returns
    -------
    None
        Updates mission segment analyses directly

    See Also
    --------
    RCAIDE.Library.Methods.Geometry.Planform
    """
    
        
    last_tag = None
    for tag,segment in mission.segments.items():  
        if segment.analyses.aerodynamics != None:
            # ensure all properties of wing are computed before drag calculations  
            vehicle =  segment.analyses.aerodynamics.vehicle
            for wing in  vehicle.wings:
                if len(wing.segments) > 1: 
                    wing_segmented_planform(wing)
                else:
                    wing_planform(wing)
                
            if (last_tag!=  None) and  ('compute' in mission.segments[last_tag].analyses.aerodynamics.process.keys()): 
                segment.analyses.aerodynamics.process.compute.lift.inviscid_wings = mission.segments[last_tag].analyses.aerodynamics.process.compute.lift.inviscid_wings
                segment.analyses.aerodynamics.surrogates       = mission.segments[last_tag].analyses.aerodynamics.surrogates 
                segment.analyses.aerodynamics.reference_values = mission.segments[last_tag].analyses.aerodynamics.reference_values  
            else:          
                aero   = segment.analyses.aerodynamics
                aero.initialize()   
                last_tag = tag  
    return 


# RCAIDE/Library/Missions/Segments/Hover/Hover.py
# 
# 
# Created:  Jul 2023, M. Clarke 
 
# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------  

def initialize_conditions(segment):
    """"
    Initializes conditions for hover segment at fixed altitude

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed
            - altitude : float
                Hover altitude [m]
            - time : float
                Hover duration [s]
            - state:
                numerics:
                    dimensionless:
                        control_points : array
                            Discretization points [-]
                conditions:
                    frames:
                        inertial:
                            time : array
                                Time points [s]
                initials : Data, optional
                    Initial conditions from previous segment
    
    Returns
    -------
    None
        Updates segment conditions directly:
            - conditions.freestream.altitude [m]
            - conditions.frames.inertial.position_vector [m]
            - conditions.frames.inertial.velocity_vector [m/s]
            - conditions.frames.inertial.time [s]

    Notes
    -----
    This function sets up the initial conditions for a hover segment with fixed
    altitude and duration. The segment represents stationary flight with no
    translational velocity components.

    **Calculation Process**
        1. Check initial altitude
        2. Scale time points over hover duration:
            t = t_norm * duration + t_initial
        3. Set velocity components to zero:
            - v_x = 0
            - v_y = 0
            - v_z = 0
        4. Set position at hover altitude

    **Major Assumptions**
        * Perfect hover (no drift)
        * Fixed altitude
        * No translational velocity
        * No atmospheric variations
        * Quasi-steady state

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    """       
    
    # unpack
    alt        = segment.altitude
    duration   = segment.time
    conditions = segment.state.conditions   
    
    
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2]      
    
    # dimensionalize time
    t_initial = conditions.frames.inertial.time[0,0]
    t_nondim  = segment.state.numerics.dimensionless.control_points
    time      =  t_nondim * (duration) + t_initial
    
    # pack
    segment.state.conditions.freestream.altitude[:,0]             = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down
    segment.state.conditions.frames.inertial.velocity_vector[:,0] = 0.
    segment.state.conditions.frames.inertial.time[:,0]            = time[:,0]    



# RCAIDE/Library/Methods/Mass_Properties/Weight_Buildups/Conventional/General_Aviation/FLOPS/__init__.py
# 

"""RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from .compute_fuselage_weight import compute_fuselage_weight
from .compute_horizontal_tail_weight import compute_horizontal_tail_weight
from .compute_vertical_tail_weight import compute_vertical_tail_weight
from .compute_operating_empty_weight import compute_operating_empty_weight
from .compute_propulsion_system_weight import compute_propulsion_system_weight
from .compute_systems_weight import compute_systems_weight
from .compute_wing_weight import compute_wing_weight
from .compute_landing_gear_weight import compute_landing_gear_weight
from .compute_operating_items_weight import compute_operating_items_weight


# RCAIDE/Methods/Energy/Propulsors/Networks/Turbojet/compute_turbojet_performance.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports  
from RCAIDE.Framework.Core import Data    
from RCAIDE.Library.Methods.Powertrain.Converters.Ram                import compute_ram_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Combustor          import compute_combustor_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Compressor         import compute_compressor_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Turbine            import compute_turbine_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Supersonic_Nozzle  import compute_supersonic_nozzle_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Compression_Nozzle import compute_compression_nozzle_performance
from RCAIDE.Library.Methods.Powertrain.Propulsors.Turbojet           import compute_thrust

# python imports 
import  numpy as  np 
from copy import  deepcopy

# ----------------------------------------------------------------------------------------------------------------------
# compute_turbojet_performance
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_turbojet_performance(turbojet, state, center_of_gravity=[[0.0, 0.0, 0.0]]):
    """
    Computes the performance of a turbojet engine by analyzing the thermodynamic cycle.
    
    Parameters
    ----------
    turbojet : RCAIDE.Library.Components.Propulsors.Turbojet
        Turbojet engine component with the following attributes:
            - tag : str
                Identifier for the turbojet
            - working_fluid : Data
                Working fluid properties object
            - ram : Data
                Ram component
                    - tag : str
                        Identifier for the ram
            - inlet_nozzle : Data
                Inlet nozzle component
                    - tag : str
                        Identifier for the inlet nozzle
            - low_pressure_compressor : Data
                Low pressure compressor component
                    - tag : str
                        Identifier for the low pressure compressor
                    - motor : Data, optional
                        Electric motor component
                    - generator : Data, optional
                        Electric generator component
                    - design_angular_velocity : float
                        Design angular velocity [rad/s]
            - high_pressure_compressor : Data
                High pressure compressor component
                    - tag : str
                        Identifier for the high pressure compressor
                    - design_angular_velocity : float
                        Design angular velocity [rad/s]
            - combustor : Data
                Combustor component
                    - tag : str
                        Identifier for the combustor
                    - fuel_data : Data
                        Fuel properties
                        - specific_energy : float
                            Fuel specific energy [J/kg]
            - high_pressure_turbine : Data
                High pressure turbine component
                    - tag : str
                        Identifier for the high pressure turbine
            - low_pressure_turbine : Data
                Low pressure turbine component
                    - tag : str
                        Identifier for the low pressure turbine
            - afterburner : Data
                Afterburner component
                    - tag : str
                        Identifier for the afterburner
            - core_nozzle : Data
                Core nozzle component
                    - tag : str
                        Identifier for the core nozzle
            - afterburner_active : bool
                Flag indicating if afterburner is active
            - origin : list of lists
                Origin coordinates [[x, y, z]] [m]
    state : RCAIDE.Framework.Mission.Common.State
        State object containing:
            - conditions : Data
                Flight conditions
                    - freestream : Data
                        Freestream properties
                            - velocity : numpy.ndarray
                                Freestream velocity [m/s]
                            - temperature : numpy.ndarray
                                Freestream temperature [K]
                            - pressure : numpy.ndarray
                                Freestream pressure [Pa]
                    - noise : Data
                        Noise conditions
                            - propulsors : dict
                                Propulsor noise conditions indexed by tag
                    - energy : Data
                        Energy conditions
                            - propulsors : dict
                                Propulsor energy conditions indexed by tag
                            - converters : dict
                                Converter energy conditions indexed by tag
                            - hybrid_power_split_ratio : float
                                Ratio of power split for hybrid systems
            - numerics : Data
                Numerical properties
                    - time : Data
                        Time properties
                            - differentiate : list
                                List of differentiation methods
            - ones_row : function
                Function to create array of ones with specified length
    center_of_gravity : list of lists, optional
        Center of gravity coordinates [[x, y, z]] [m]. Default: [[0.0, 0.0, 0.0]]
    
    Returns
    -------
    thrust_vector : numpy.ndarray
        Thrust force vector [N]
    moment : numpy.ndarray
        Moment vector [N·m]
    power : numpy.ndarray
        Shaft power output [W]
    power_elec : numpy.ndarray
        Electrical power input/output [W]
    stored_results_flag : bool
        Flag indicating if results are stored
    stored_propulsor_tag : str
        Tag of the turbojet with stored results
    
    Notes
    -----
    This function computes the performance of a turbojet engine by sequentially analyzing
    each component in the engine's thermodynamic cycle. It links the output conditions of
    each component to the input conditions of the next component in the flow path.
    
    The function follows this sequence:
        1. Set working fluid properties
        2. Compute ram performance
        3. Compute inlet nozzle performance
        4. Compute low pressure compressor performance
        5. Compute high pressure compressor performance
        6. Compute combustor performance
        7. Compute high pressure turbine performance
        8. Compute low pressure turbine performance
        9. Compute afterburner performance (if active)
        10. Compute core nozzle performance
        11. Compute thrust and power output
        12. Calculate efficiencies
        13. Handle electrical power generation/consumption if applicable
    
    **Major Assumptions**
        * Steady state operation
        * One-dimensional flow through components
        * Adiabatic components except for the combustor and afterburner
        * Perfect gas behavior with variable properties
    
    References
    ----------
    [1] Mattingly, J.D., "Elements of Gas Turbine Propulsion", 2nd Edition, AIAA Education Series, 2005. https://soaneemrana.org/onewebmedia/ELEMENTS%20OF%20GAS%20TURBINE%20PROPULTION2.pdf
    [2] Cantwell, B., "AA283 Course Notes", Stanford University. https://web.stanford.edu/~cantwell/AA283_Course_Material/
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Propulsors.Turbojet.compute_thrust
    RCAIDE.Library.Methods.Powertrain.Propulsors.Turbojet.reuse_stored_turbojet_data
    """
    conditions                = state.conditions
    noise_conditions          = conditions.noise.propulsors[turbojet.tag]  
    turbojet_conditions       = conditions.energy.propulsors[turbojet.tag] 
    U0                        = conditions.freestream.velocity
    T                         = conditions.freestream.temperature
    P                         = conditions.freestream.pressure
    ram                       = turbojet.ram
    inlet_nozzle              = turbojet.inlet_nozzle
    low_pressure_compressor   = turbojet.low_pressure_compressor
    high_pressure_compressor  = turbojet.high_pressure_compressor
    combustor                 = turbojet.combustor
    high_pressure_turbine     = turbojet.high_pressure_turbine
    low_pressure_turbine      = turbojet.low_pressure_turbine 
    afterburner               = turbojet.afterburner 
    core_nozzle               = turbojet.core_nozzle   

    # unpack component conditions 
    ram_conditions          = conditions.energy.converters[ram.tag]     
    inlet_nozzle_conditions = conditions.energy.converters[inlet_nozzle.tag]
    core_nozzle_conditions  = conditions.energy.converters[core_nozzle.tag] 
    lpc_conditions          = conditions.energy.converters[low_pressure_compressor.tag]
    hpc_conditions          = conditions.energy.converters[high_pressure_compressor.tag]
    lpt_conditions          = conditions.energy.converters[low_pressure_turbine.tag]
    hpt_conditions          = conditions.energy.converters[high_pressure_turbine.tag]
    combustor_conditions    = conditions.energy.converters[combustor.tag]
    afterburner_conditions  = conditions.energy.converters[afterburner.tag] 
    
    # Set the working fluid to determine the fluid properties
    ram.working_fluid = turbojet.working_fluid

    # Flow through the ram , this computes the necessary flow quantities and stores it into conditions
    compute_ram_performance(ram,conditions)

    # Link inlet nozzle to ram 
    inlet_nozzle_conditions.inputs.stagnation_temperature             = ram_conditions.outputs.stagnation_temperature 
    inlet_nozzle_conditions.inputs.stagnation_pressure                = ram_conditions.outputs.stagnation_pressure
    inlet_nozzle_conditions.inputs.static_temperature                 = ram_conditions.outputs.static_temperature
    inlet_nozzle_conditions.inputs.static_pressure                    = ram_conditions.outputs.static_pressure
    inlet_nozzle_conditions.inputs.mach_number                        = ram_conditions.outputs.mach_number
    inlet_nozzle.working_fluid                                        = ram.working_fluid

    # Flow through the inlet nozzle
    compute_compression_nozzle_performance(inlet_nozzle,conditions)

    # Link low pressure compressor to the inlet nozzle
    lpc_conditions.inputs.stagnation_temperature  = inlet_nozzle_conditions.outputs.stagnation_temperature
    lpc_conditions.inputs.stagnation_pressure     = inlet_nozzle_conditions.outputs.stagnation_pressure
    lpc_conditions.inputs.static_temperature      = inlet_nozzle_conditions.outputs.static_temperature
    lpc_conditions.inputs.static_pressure         = inlet_nozzle_conditions.outputs.static_pressure
    lpc_conditions.inputs.mach_number             = inlet_nozzle_conditions.outputs.mach_number  
    low_pressure_compressor.working_fluid         = inlet_nozzle.working_fluid

    # Flow through the low pressure compressor
    compute_compressor_performance(low_pressure_compressor,conditions)

    # Link the high pressure compressor to the low pressure compressor
    hpc_conditions.inputs.stagnation_temperature = lpc_conditions.outputs.stagnation_temperature
    hpc_conditions.inputs.stagnation_pressure    = lpc_conditions.outputs.stagnation_pressure
    hpc_conditions.inputs.static_temperature     = lpc_conditions.outputs.static_temperature
    hpc_conditions.inputs.static_pressure        = lpc_conditions.outputs.static_pressure
    hpc_conditions.inputs.mach_number            = lpc_conditions.outputs.mach_number  
    high_pressure_compressor.working_fluid       = low_pressure_compressor.working_fluid    

    # Flow through the high pressure compressor
    compute_compressor_performance(high_pressure_compressor,conditions)

    # Link the combustor to the high pressure compressor 
    combustor_conditions.inputs.stagnation_temperature   = hpc_conditions.outputs.stagnation_temperature
    combustor_conditions.inputs.stagnation_pressure      = hpc_conditions.outputs.stagnation_pressure
    combustor_conditions.inputs.static_temperature       = hpc_conditions.outputs.static_temperature
    combustor_conditions.inputs.static_pressure          = hpc_conditions.outputs.static_pressure
    combustor_conditions.inputs.mach_number              = hpc_conditions.outputs.mach_number  
    combustor.working_fluid                              = high_pressure_compressor.working_fluid  

    # Flow through the combustor
    compute_combustor_performance(combustor,conditions)

    # Link the high pressure turbine to the combustor
    hpt_conditions.inputs.stagnation_temperature    = combustor_conditions.outputs.stagnation_temperature
    hpt_conditions.inputs.stagnation_pressure       = combustor_conditions.outputs.stagnation_pressure
    hpt_conditions.inputs.fuel_to_air_ratio         = combustor_conditions.outputs.fuel_to_air_ratio 
    hpt_conditions.inputs.static_temperature        = combustor_conditions.outputs.static_temperature
    hpt_conditions.inputs.static_pressure           = combustor_conditions.outputs.static_pressure
    hpt_conditions.inputs.mach_number               = combustor_conditions.outputs.mach_number  
    hpt_conditions.inputs.compressor                = hpc_conditions.outputs
    hpt_conditions.inputs.bypass_ratio              = 0.0
    high_pressure_turbine.working_fluid             = combustor.working_fluid 

    # Flow through the high pressure turbine
    compute_turbine_performance(high_pressure_turbine,conditions)

    # Link the low pressure turbine to the high pressure turbine
    lpt_conditions.inputs.stagnation_temperature     = hpt_conditions.outputs.stagnation_temperature
    lpt_conditions.inputs.stagnation_pressure        = hpt_conditions.outputs.stagnation_pressure 
    lpt_conditions.inputs.static_temperature         = hpt_conditions.outputs.static_temperature
    lpt_conditions.inputs.static_pressure            = hpt_conditions.outputs.static_pressure 
    lpt_conditions.inputs.mach_number                = hpt_conditions.outputs.mach_number    
    lpt_conditions.inputs.compressor                 = lpc_conditions.outputs
    lpt_conditions.inputs.fuel_to_air_ratio          = combustor_conditions.outputs.fuel_to_air_ratio 
    lpt_conditions.inputs.bypass_ratio               = 0.0
    low_pressure_turbine.working_fluid               = high_pressure_turbine.working_fluid    

    # Flow through the low pressure turbine
    compute_turbine_performance(low_pressure_turbine,conditions)
 

    if turbojet.afterburner_active == True:
        #link the core nozzle to the afterburner
        afterburner_conditions.inputs.stagnation_temperature = lpt_conditions.outputs.stagnation_temperature
        afterburner_conditions.inputs.stagnation_pressure    = lpt_conditions.outputs.stagnation_pressure   
        afterburner_conditions.inputs.nondim_ratio           = 1.0 + combustor_conditions.outputs.fuel_to_air_ratio
        afterburner_conditions.inputs.static_temperature     = lpt_conditions.outputs.static_temperature
        afterburner_conditions.inputs.static_pressure        = lpt_conditions.outputs.static_pressure
        afterburner_conditions.inputs.mach_number            = lpt_conditions.outputs.mach_number  
        afterburner.working_fluid                            = low_pressure_turbine.working_fluid

        #flow through the afterburner 
        compute_combustor_performance(afterburner,conditions)

        #link the core nozzle to the afterburner
        core_nozzle_conditions.inputs.stagnation_temperature = afterburner_conditions.outputs.stagnation_temperature
        core_nozzle_conditions.inputs.stagnation_pressure    = afterburner_conditions.outputs.stagnation_pressure  
        core_nozzle_conditions.inputs.static_temperature     = afterburner_conditions.outputs.static_temperature
        core_nozzle_conditions.inputs.static_pressure        = afterburner_conditions.outputs.static_pressure  
        core_nozzle_conditions.inputs.mach_number            = afterburner_conditions.outputs.mach_number   
        core_nozzle.working_fluid                            = afterburner.working_fluid  

    else:
        #link the core nozzle to the low pressure turbine
        core_nozzle_conditions.inputs.stagnation_temperature = lpt_conditions.outputs.stagnation_temperature
        core_nozzle_conditions.inputs.stagnation_pressure    = lpt_conditions.outputs.stagnation_pressure
        core_nozzle_conditions.inputs.static_temperature     = lpt_conditions.outputs.static_temperature
        core_nozzle_conditions.inputs.static_pressure        = lpt_conditions.outputs.static_pressure  
        core_nozzle_conditions.inputs.mach_number            = lpt_conditions.outputs.mach_number   
        core_nozzle.working_fluid                            = low_pressure_compressor.working_fluid 
 
    # Flow through the core nozzle
    compute_supersonic_nozzle_performance(core_nozzle,conditions) 
 
    # Link the thrust component to the core nozzle 
    turbojet_conditions.core_nozzle_area_ratio                          = core_nozzle_conditions.outputs.area_ratio 
    turbojet_conditions.core_nozzle_static_pressure                     = core_nozzle_conditions.outputs.static_pressure
    turbojet_conditions.core_nozzle_exit_velocity                       = core_nozzle_conditions.outputs.velocity  

    # Link the thrust component to the combustor
    turbojet_conditions.fuel_to_air_ratio                        = combustor_conditions.outputs.fuel_to_air_ratio 
    if turbojet.afterburner_active == True:
        # previous fuel ratio is neglected when the afterburner fuel ratio is calculated
        turbojet_conditions.fuel_to_air_ratio += afterburner_conditions.outputs.fuel_to_air_ratio

    # Link the thrust component to the low pressure compressor 
    turbojet_conditions.total_temperature_reference              = lpc_conditions.outputs.stagnation_temperature
    turbojet_conditions.total_pressure_reference                 = lpc_conditions.outputs.stagnation_pressure 
    turbojet_conditions.flow_through_core                        = 1.0 #scaled constant to turn on core thrust computation  
    
    # Compute the thrust
    compute_thrust(turbojet,conditions)
    
    # Compute forces and moments
    moment_vector              = 0*state.ones_row(3)
    thrust_vector              = 0*state.ones_row(3)
    thrust_vector[:,0]         = turbojet_conditions.thrust[:,0]
    moment_vector[:,0]         = turbojet.origin[0][0] -   center_of_gravity[0][0] 
    moment_vector[:,1]         = turbojet.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2]         = turbojet.origin[0][2]  -  center_of_gravity[0][2]
    M                          = np.cross(moment_vector, thrust_vector)   
    moment                     = M 
    power                      = turbojet_conditions.power 
    turbojet_conditions.moment = moment 

    # compute efficiencies 
    mdot_air_core                                  = turbojet_conditions.core_mass_flow_rate 
    fuel_enthalpy                                  = combustor.fuel_data.specific_energy 
    mdot_fuel                                      = turbojet_conditions.fuel_flow_rate   
    h_e_c                                          = core_nozzle_conditions.outputs.static_enthalpy
    h_0                                            = turbojet.working_fluid.compute_cp(T,P) * T 
    h_t4                                           = combustor_conditions.outputs.stagnation_enthalpy
    h_t3                                           = hpc_conditions.outputs.stagnation_enthalpy 
    turbojet_conditions.overall_efficiency         = thrust_vector* U0 / (mdot_fuel * fuel_enthalpy)  
    turbojet_conditions.thermal_efficiency         = 1 - ((mdot_air_core +  mdot_fuel)*(h_e_c -  h_0) + mdot_fuel *h_0)/((mdot_air_core +  mdot_fuel)*h_t4 - mdot_air_core *h_t3)  
 

    # compute shaft RPMs   
    lpc_conditions.omega        = low_pressure_compressor.design_angular_velocity * turbojet_conditions.throttle
    hpc_conditions.omega        = high_pressure_compressor.design_angular_velocity * turbojet_conditions.throttle
   
    # compute electrical power if generated/supplied  
    power_elec = 0*state.ones_row(1)
    if low_pressure_compressor.motor != None and  len(state.numerics.time.differentiate) > 0: 
        compressor_motor_conditions                 = conditions.energy.converters[low_pressure_compressor.motor.tag] 
        compressor_motor_conditions.outputs.power   = power *conditions.energy.hybrid_power_split_ratio   
        compressor_motor_conditions.outputs.omega   = lpc_conditions.omega
        compressor_motor_conditions.outputs.torque  = compressor_motor_conditions.outputs.power / compressor_motor_conditions.outputs.omega   
        power_elec =  compressor_motor_conditions.outputs.power  
    
    if low_pressure_compressor.generator != None and len(state.numerics.time.differentiate) > 0: 
        compressor_generator_conditions                = conditions.energy.converters[low_pressure_compressor.generator.tag] 
        compressor_generator_conditions.inputs.power   = power *conditions.energy.hybrid_power_split_ratio  
        compressor_generator_conditions.inputs.omega   = lpc_conditions.omega
        compressor_generator_conditions.outputs.torque = compressor_generator_conditions.outputs.power / compressor_generator_conditions.outputs.omega  
        power_elec =  compressor_generator_conditions.inputs.power  

        
    # store data
    core_nozzle_res = Data(
                exit_static_temperature             = core_nozzle_conditions.outputs.static_temperature,
                exit_static_pressure                = core_nozzle_conditions.outputs.static_pressure,
                exit_stagnation_temperature         = core_nozzle_conditions.outputs.stagnation_temperature,
                exit_stagnation_pressure            = core_nozzle_conditions.outputs.static_pressure,
                exit_velocity                       = core_nozzle_conditions.outputs.velocity
            )

    lpc_res = Data(
                angular_velocity    =  lpc_conditions.omega, 
            )    

    noise_conditions.fan_nozzle             = None 
    noise_conditions.core_nozzle            = core_nozzle_res
    noise_conditions.fan                    = lpc_res   
    stored_results_flag                     = True
    stored_propulsor_tag                    = turbojet.tag
    
    power_elec =  0*state.ones_row(1)
    
    return thrust_vector,moment,power,power_elec,stored_results_flag,stored_propulsor_tag 

def reuse_stored_turbojet_data(turbojet,state,network,stored_propulsor_tag,center_of_gravity= [[0.0, 0.0,0.0]]):
    '''Reuses results from one turbojet for identical propulsors
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    turbojet            - turbojet data structure                [-] 
    state               - operating conditions data structure   [-]  
    fuel_line            - fuelline                              [-] 
    total_thrust         - thrust of turbojet group              [N]
    total_power          - power of turbojet group               [W] 

    Outputs:  
    total_thrust         - thrust of turbojet group              [N]
    total_power          - power of turbojet group               [W] 
    
    Properties Used: 
    N.A.        
    '''
    # unpack
    conditions                  = state.conditions 
    ram                         = turbojet.ram
    inlet_nozzle                = turbojet.inlet_nozzle 
    low_pressure_compressor     = turbojet.low_pressure_compressor
    high_pressure_compressor    = turbojet.high_pressure_compressor
    combustor                   = turbojet.combustor
    high_pressure_turbine       = turbojet.high_pressure_turbine
    low_pressure_turbine        = turbojet.low_pressure_turbine
    core_nozzle                 = turbojet.core_nozzle
    ram_0                       = network.propulsors[stored_propulsor_tag].ram
    inlet_nozzle_0              = network.propulsors[stored_propulsor_tag].inlet_nozzle 
    low_pressure_compressor_0   = network.propulsors[stored_propulsor_tag].low_pressure_compressor
    high_pressure_compressor_0  = network.propulsors[stored_propulsor_tag].high_pressure_compressor
    combustor_0                 = network.propulsors[stored_propulsor_tag].combustor
    high_pressure_turbine_0     = network.propulsors[stored_propulsor_tag].high_pressure_turbine
    low_pressure_turbine_0      = network.propulsors[stored_propulsor_tag].low_pressure_turbine
    core_nozzle_0               = network.propulsors[stored_propulsor_tag].core_nozzle

    # deep copy results 
    conditions.energy.propulsors[turbojet.tag]                 = deepcopy(conditions.energy.propulsors[stored_propulsor_tag])
    conditions.noise.propulsors[turbojet.tag]                  = deepcopy(conditions.noise.propulsors[stored_propulsor_tag]) 
    conditions.energy.converters[ram.tag]                      = deepcopy(conditions.energy.converters[ram_0.tag]                     )
    conditions.energy.converters[inlet_nozzle.tag]             = deepcopy(conditions.energy.converters[inlet_nozzle_0.tag]            ) 
    conditions.energy.converters[low_pressure_compressor.tag]  = deepcopy(conditions.energy.converters[low_pressure_compressor_0.tag] )
    conditions.energy.converters[high_pressure_compressor.tag] = deepcopy(conditions.energy.converters[high_pressure_compressor_0.tag])
    conditions.energy.converters[combustor.tag]                = deepcopy(conditions.energy.converters[combustor_0.tag]               )
    conditions.energy.converters[low_pressure_turbine.tag]     = deepcopy(conditions.energy.converters[low_pressure_turbine_0.tag]    )
    conditions.energy.converters[high_pressure_turbine.tag]    = deepcopy(conditions.energy.converters[high_pressure_turbine_0.tag]   )
    conditions.energy.converters[core_nozzle.tag]              = deepcopy(conditions.energy.converters[core_nozzle_0.tag]             )

    # compute moment  
    moment_vector      = 0*state.ones_row(3)
    thrust_vector      = 0*state.ones_row(3)
    thrust_vector[:,0] = conditions.energy.propulsors[turbojet.tag].thrust[:,0] 
    moment_vector[:,0] = turbojet.origin[0][0] -   center_of_gravity[0][0] 
    moment_vector[:,1] = turbojet.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2] = turbojet.origin[0][2]  -  center_of_gravity[0][2]
    moment             = np.cross(moment_vector,thrust_vector)    
  
    power                                             = conditions.energy.propulsors[turbojet.tag].power 
    conditions.energy.propulsors[turbojet.tag].moment = moment

    power_elec = 0*state.ones_row(1)
    if low_pressure_compressor.motor != None and  len(state.numerics.time.differentiate) > 0: 
        conditions.energy.converters[low_pressure_compressor.motor.tag]  = deepcopy(conditions.energy.converters[low_pressure_compressor_0.motor.tag]) 
        power_elec =  conditions.energy.converters[low_pressure_compressor.motor.tag].outputs.power  
    
    if low_pressure_compressor.generator != None and len(state.numerics.time.differentiate) > 0:  
        conditions.energy.converters[low_pressure_compressor.generator.tag]  = deepcopy(conditions.energy.converters[low_pressure_compressor_0.generator.tag]) 
        power_elec =  conditions.energy.converters[low_pressure_compressor.generator.tag].inputs.power
        
    return thrust_vector,moment,power, power_elec
 


# RCAIDE/Library/Methods/Geometry/LOPA/LOPA_functions.py
# 
# 
# Created: Mar 2025, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports 
import RCAIDE 
from .LOPA_functions import *

# python functions 
import  numpy as  np 
from copy import  deepcopy

# ----------------------------------------------------------------------------------------------------------------------
#  compute_layout_of_passenger_accommodations
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_layout_of_passenger_accommodations(fuselage):
    '''
    Creates the layout of passenger accommodations for a vehicle
    '''  
    #  [ x, y, z , length, width, first-cl flag, business-cl flag, economy-cl flag, seat, emergency-row flag, galley/lav flag, type-A exit flag]    
    LOPA = np.empty(( 0, 14)) 
    side_cabin_offset = 0
    for cabin in fuselage.cabins: 
        cabin_class_origin  = [0, 0, 0]
        for cabin_class in cabin.classes: 
            seat_data ,cabin_class_origin  = create_class_seating_map_layout(cabin, cabin_class,cabin_class_origin, side_cabin_offset)
            side_cabin_offset =  cabin.width / 2
            LOPA = np.vstack((LOPA,seat_data))  
            
    for cabin in fuselage.cabins: 
        for cabin_class in cabin.classes: 
            cabin_class.percentage = cabin_class.length/cabin.length

    fuselage.number_of_passengers               = np.sum(LOPA[:,10])
    fuselage.layout_of_passenger_accommodations = LOPA
    compute_fusselage_dimensions(fuselage)

    return  
 
def compute_fusselage_dimensions(fuselage):
    LOPA = fuselage.layout_of_passenger_accommodations 
    
       # Step 1: plot cabin bounds  
    # get points at x min 
    x_min_locs   =  np.where( LOPA[:,2] == min(LOPA[:,2]))[0]
    x_min        =  LOPA[x_min_locs[0],2] -  LOPA[x_min_locs[0],5]/2
    x_min_y_max  =  max(LOPA[x_min_locs,3] + LOPA[x_min_locs,6]/2 )
    x_min_y_min  =  min(LOPA[x_min_locs,3] - LOPA[x_min_locs,6]/2 ) 
    x_border_pts = [x_min, x_min] 
    y_border_pts = [x_min_y_min, x_min_y_max] 

    # get points at y max 
    y_max_locs   =  np.where( LOPA[:,3] == max(LOPA[:,3]))[0]
    y_max        =  LOPA[y_max_locs[0],3] + LOPA[y_max_locs[0],6]/2 
    y_max_x_max  =  max(LOPA[y_max_locs,2] + LOPA[y_max_locs[0],5]/2)
    y_max_x_min  =  min(LOPA[y_max_locs,2] - LOPA[y_max_locs[0],5]/2) 
    x_border_pts.append(y_max_x_min)
    x_border_pts.append(y_max_x_max)
    y_border_pts.append(y_max)
    y_border_pts.append(y_max) 

    # get points at x max 
    x_max_locs   =  np.where( LOPA[:,2] == max(LOPA[:,2]))[0]
    x_max        =  LOPA[x_max_locs[0],2] + LOPA[x_max_locs[0],5]/2
    x_max_y_max  =  max(LOPA[x_max_locs,3] + LOPA[x_max_locs,6]/2)
    x_max_y_min  =  min(LOPA[x_max_locs,3] - LOPA[x_max_locs,6]/2)  
    x_border_pts.append(x_max)
    x_border_pts.append(x_max)
    y_border_pts.append(x_max_y_max)
    y_border_pts.append(x_max_y_min)
    
    # get points at y min  
    y_min_locs   =  np.where( LOPA[:,3] == min(LOPA[:,3]))[0]
    y_min        =  LOPA[y_min_locs[0],3] - LOPA[y_min_locs[0],6]/2 
    y_min_x_max  =  max(LOPA[y_min_locs,2] + LOPA[y_min_locs[0],5]/2)
    y_min_x_min  =  min(LOPA[y_min_locs,2] - LOPA[y_min_locs[0],5]/2)
    x_border_pts.append(y_min_x_max)  
    x_border_pts.append(y_min_x_min)
    y_border_pts.append(y_min)
    y_border_pts.append(y_min)    
    
    # loop through points and determine if there are duplicates
    y_border_pts = np.array(y_border_pts)
    x_border_pts = np.array(x_border_pts)

    # cut where y is negative
    port_idxs  =  np.where(y_border_pts<0)[0]
    starboard_x_points = np.delete(x_border_pts, port_idxs) 
    starboard_y_points = np.delete(y_border_pts, port_idxs)
    
    cabin_length = max(starboard_x_points)
    cabin_wdith = 2*max(starboard_y_points)

    fuselage.lengths.nose = fuselage.fineness.nose*cabin_wdith
    fuselage.lengths.tail = fuselage.fineness.tail*cabin_wdith  

    fuselage.lengths.total = fuselage.lengths.nose + fuselage.lengths.tail + cabin_length
    fuselage.width  = cabin_wdith

def create_class_seating_map_layout(cabin,cabin_class,cabin_class_origin, side_cabin_offset):  
    
    s_y_coord, cabin_class_origin = get_seat_y_coords(cabin, cabin_class,cabin_class_origin)
    s_x_coord,object_type, cabin_class_origin = get_seat_x_coords(cabin, cabin_class,cabin_class_origin) 
         
    # concatenate arrays 
    length   =  cabin_class.seat_length * np.ones_like(s_x_coord) 
    length[object_type[:,2] == 1] = cabin.galley_lavatory_length
    length[object_type[:,3] == 1] = cabin.type_A_door_length
    
    X_coords   = np.atleast_2d((np.tile(s_x_coord[:,None], (1, len(s_y_coord)))).flatten()).T
    Y_coords   = np.atleast_2d((np.tile(s_y_coord[None,:], (len(s_x_coord), 1))).flatten()).T
    Z_coords   = np.atleast_2d((np.zeros_like(Y_coords)).flatten()).T 
    length     = np.atleast_2d((np.tile(length[:,None], (1, len(s_y_coord)))).flatten()).T 
    width      = cabin_class.seat_width * np.ones_like(Z_coords) 
    n_rows     = cabin_class.number_of_rows * np.ones_like(Z_coords) 
    n_seats_y  = cabin_class.number_of_seats_abrest  * np.ones_like(Z_coords) 
    object_vec = np.repeat(object_type, len(s_y_coord), axis=0)
     
    # cabin class flags 
    F_c  =  np.zeros_like(Z_coords)
    B_c  =  np.zeros_like(Z_coords)
    E_c  =  np.zeros_like(Z_coords)     
    if type(cabin_class) == RCAIDE.Library.Components.Fuselages.Cabins.Classes.First: 
        F_c  =  np.ones_like(Z_coords) 
    elif type(cabin_class) == RCAIDE.Library.Components.Fuselages.Cabins.Classes.Business: 
        B_c  =  np.ones_like(Z_coords) 
    elif type(cabin_class) == RCAIDE.Library.Components.Fuselages.Cabins.Classes.Economy: 
        E_c  =  np.ones_like(Z_coords)  
    
    # [n_rows , n_seats_y, x, y, z , length, width, first-cl flag, business-cl flag, economy-cl flag, seat, emergency-row flag, galley/lav flag, type-A exit flag]    
    seat_data = np.hstack((n_rows , n_seats_y, X_coords,Y_coords, Z_coords, length ,width,F_c,B_c,E_c,object_vec))
     
    if type(cabin) == RCAIDE.Library.Components.Fuselages.Cabins.Side_Cabin:
        
        seat_data[:, 3] += cabin.width / 2
        seat_data  = update_seat_map_layout_using_cabin_taper(seat_data,cabin) 
        seat_data[:, 3] += side_cabin_offset  
        
        # make copy about center
        seat_data_        = deepcopy(seat_data)
        seat_data_[:, 3] *= -1 
        seat_data         = np.vstack((seat_data,seat_data_))
        
    return seat_data ,cabin_class_origin 


    
    


# RCAIDE/Methods/Energy/Converters/Turboshaft/size_core.py
# 
# 
# Created:  Jul 2023, M. Clarke
# Modified: Jun 2024, M. Guidotti  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from RCAIDE.Library.Methods.Powertrain.Converters.Turboshaft import compute_power

# Python package imports
import numpy                                                       as np

# ----------------------------------------------------------------------------------------------------------------------
#  size_core
# ----------------------------------------------------------------------------------------------------------------------
def size_core(turboshaft, conditions):
    """
    Sizes the core flow for a turboshaft engine at the design condition.
    
    Parameters
    ----------
    turboshaft : RCAIDE.Library.Components.Converters.Turboshaft
        Turboshaft engine component with the following attributes:
            - tag : str
                Identifier for the turboshaft
            - reference_temperature : float
                Reference temperature [K]
            - reference_pressure : float
                Reference pressure [Pa]
            - design_power : float
                Design power output [W]
            - compressor : Data
                Compressor component
                    - mass_flow_rate : float
                        Mass flow rate through the compressor [kg/s]
            - mass_flow_rate : float
                Mass flow rate through the turboshaft [kg/s]
    conditions : RCAIDE.Framework.Mission.Common.Conditions
        Flight conditions with:
            - energy : Data
                Energy conditions
                    - converters : dict
                        Converter energy conditions indexed by tag
                            - total_temperature_reference : float
                                Reference total temperature [K]
                            - total_pressure_reference : float
                                Reference total pressure [Pa]
                            - non_dimensional_power : float
                                Non-dimensional power
    
    Returns
    -------
    None
        Results are stored in the turboshaft object:
            - mass_flow_rate : float
                Mass flow rate through the turboshaft [kg/s]
            - compressor.mass_flow_rate : float
                Mass flow rate through the compressor [kg/s]
    
    Notes
    -----
    This function sizes the core flow of a turboshaft engine based on the design power
    requirement and the non-dimensional power computed from the engine cycle analysis.
    It calculates the dimensional mass flow rate needed to produce the specified design
    power and corrects it to the reference conditions.
    
    The computation follows these steps:
        1. Extract reference conditions and engine parameters
        2. Compute non-dimensional power using the compute_power function
        3. Calculate the dimensional mass flow rate required to produce the design power
        4. Correct the mass flow rate to the reference conditions
        5. Store the results in the turboshaft object
    
    **Major Assumptions**
        * Perfect gas behavior
        * Turboshaft engine with free power turbine
    
    **Theory**
    The dimensional mass flow rate is calculated as:
    
    .. math::
        \\dot{m}_{air} = \\frac{P_{design}}{P_{sp}}
    
    The corrected mass flow rate for the compressor is:
    
    .. math::
        \\dot{m}_{compressor} = \\frac{\\dot{m}_{air}}{\\sqrt{\\frac{T_{ref}}{T_{t,ref}}}\\frac{P_{t,ref}}{P_{ref}}}
    
    where:
      - :math:`P_{design}` is the design power
      - :math:`P_{sp}` is the non-dimensional power
      - :math:`T_{ref}` is the reference temperature
      - :math:`T_{t,ref}` is the reference total temperature
      - :math:`P_{ref}` is the reference pressure
      - :math:`P_{t,ref}` is the reference total pressure
      
    References
    ----------
    [1] Mattingly, J.D., "Elements of Gas Turbine Propulsion", AIAA Education Series, 2005, pp. 332-336
    [2] Stuyvenberg, L., "Helicopter Turboshafts", University of Colorado, 2015 https://www.colorado.edu/faculty/kantha/sites/default/files/attached-files/70652-116619_-_luke_stuyvenberg_-_dec_17_2015_1258_pm_-_stuyvenberg_helicopterturboshafts.pdf
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Turboshaft.compute_power
    """
    
    #unpack from turboshaft
    turboshaft_conditions                          = conditions.energy.converters[turboshaft.tag] 
    Tref                                           = turboshaft.reference_temperature
    Pref                                           = turboshaft.reference_pressure 
    total_temperature_reference                    = turboshaft_conditions.total_temperature_reference  
    total_pressure_reference                       = turboshaft_conditions.total_pressure_reference 

    #compute nondimensional power
    compute_power(turboshaft,conditions)

    #unpack results 
    Psp                                            = turboshaft_conditions.non_dimensional_power
    
    #compute dimensional mass flow rates
    mdot_air                                       = turboshaft.design_power/Psp
    mdot_compressor                                = mdot_air/ (np.sqrt(Tref/total_temperature_reference)*(total_pressure_reference/Pref))

    #pack outputs
    turboshaft.mass_flow_rate                      = mdot_air
    turboshaft.compressor.mass_flow_rate           = mdot_compressor

    return    



# RCAIDE/Library/Missions/Common/helper_functions.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports 
import RCAIDE 
from tqdm import tqdm
import  numpy as  np

def sequential_segments(mission):

    print('\n Mission Solver Initiated')    
    pbar = tqdm(total=100)
    progress_interval = round(np.ceil(100/ len(mission.segments)))
    last_tag = None
    for tag,segment in mission.segments.items(): 
        print('\n Solving', segment.tag , 'segment.')        
        segment.mission_tag = mission.tag
        if last_tag:
            segment.state.initials = mission.segments[last_tag].state
        last_tag = tag        
        
        segment.process.initialize.expand_state(segment) 
        segment.process.initialize.expand_state = RCAIDE.Library.Methods.skip        
        segment.evaluate() 
        pbar.update(progress_interval)
        print('\n')
    pbar.close()
            
            


# RCAIDE/Methods/Mission/Segments/Transition/__init__.py
# 

"""
Collection of mission segment types that handle aircraft transition maneuvers, including constant acceleration transitions with various climb and pitch configurations.

See Also
--------
RCAIDE.Library.Mission.Segments.Cruise
RCAIDE.Library.Mission.Segments.Climb
RCAIDE.Library.Mission.Segments.Descent
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from . import Constant_Acceleration_Constant_Pitchrate_Constant_Altitude
from . import Constant_Acceleration_Constant_Angle_Linear_Climb 


# RCAIDE/Library/Methods/Powertrain/__init__.py
# 

"""
This module provides functionality for modeling and analyzing powertrain systems in aerospace vehicles. 
It includes methods for setting up operating conditions, and contains submodules for various powertrain 
components such as converters, distributors, modulators, propulsors, and energy sources.

The powertrain module enables the simulation and analysis of complex energy conversion and distribution 
systems, from energy sources through power conversion to propulsion elements.

See Also
--------
RCAIDE.Library.Components.Powertrain
RCAIDE.Library.Methods.Aerodynamics
RCAIDE.Library.Methods.Powertrain
RCAIDE.Library.Methods.Mass_Properties
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .setup_operating_conditions     import setup_operating_conditions

from . import Converters
from . import Distributors
from . import Modulators
from . import Propulsors
from . import Sources
from . import Systems


# RCAIDE/Library/Plots/__init__.py
# 

"""
This package contains modules for visualizing various aspects of aircraft analysis, simulation results,
and aircraft geometry.

See Also
--------
RCAIDE.Library.Analysis : Analysis tools generating data for plots
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Geometry           import *  
from .Aerodynamics       import * 
from .Common             import *
from .Emissions          import *  
from .Energy             import * 
from .Mission            import *  
from .Noise              import *
from .Thermal_Management import * 
from .Topography         import * 
from .Stability          import * 
from .Weights            import *


# RCAIDE/Methods/Powertrain/Converters/Fuel_Cells/Common/compute_stack_properties.py
# 
# 
# Created:   Nov 2024, M. Clarke 
# Modified:  Jan 2025, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
import RCAIDE
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Methods.Powertrain.Converters.Fuel_Cells.Larminie_Model import  compute_power, compute_voltage
from RCAIDE.Library.Methods.Powertrain.Converters.Fuel_Cells.Proton_Exchange_Membrane.compute_fuel_cell_performance import  evaluate_PEM ,  evaluate_max_gross_power, set_rated_current_density 
import  scipy as  sp
import  numpy as  np
import matplotlib.pyplot as  plt
import matplotlib.cm as cm

# ----------------------------------------------------------------------------------------------------------------------
#  Compute Stack Properties
# ----------------------------------------------------------------------------------------------------------------------  
def compute_stack_properties(fuel_cell_stack):
    """
    Calculates fuel cell stack properties based on individual cell characteristics and stack configuration.
    
    Parameters
    ----------
    fuel_cell_stack : RCAIDE.Components.Powertrain.Converters.Fuel_Cell_Stack
        The fuel cell stack component for which properties are being computed
        
    Returns
    -------
    None
    
    Notes
    -----
    This function computes geometric, electrical, and performance properties of a fuel cell stack
    based on individual cell properties and stack configuration. It handles both generic fuel cell
    stacks and proton exchange membrane (PEM) fuel cell stacks with different calculation methods.
    
    For generic fuel cells, optimization is used to find the maximum power point.
    For PEM fuel cells, specialized performance models are used that account for
    operating conditions like temperature and pressure.
    
    The function calculates:
        - Physical dimensions (length, width, height) based on cell arrangement
        - Maximum power, voltage, and current
        - Fuel consumption rates
        - Mass properties
    
    **Major Assumptions**
        * Cell properties are uniform across the stack
        * Cells are arranged in a regular geometric pattern
        * For PEM cells, standard atmospheric conditions are used if not specified
        * Thermal effects within the stack are simplified
    
    **Theory**
    
    For generic fuel cells, the maximum power point is found by optimizing:
    
    .. math::
        P_{max} = \\max_{j \\in [j_{min}, j_{max}]} P(j)
    
    where j is the current density and P is the power function.
    
    For PEM fuel cells, detailed electrochemical models account for:
        - Activation losses
        - Ohmic losses
        - Concentration losses
        
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Fuel_Cells.Larminie_Model.compute_power
    RCAIDE.Library.Methods.Powertrain.Converters.Fuel_Cells.Proton_Exchange_Membrane.compute_fuel_cell_performance
    """
     
    
    series_e           = fuel_cell_stack.electrical_configuration.series
    parallel_e         = fuel_cell_stack.electrical_configuration.parallel
    n_total            = parallel_e *series_e
    normal_count       = fuel_cell_stack.geometrtic_configuration.normal_count  
    parallel_count     = fuel_cell_stack.geometrtic_configuration.parallel_count
    stacking_rows      = fuel_cell_stack.geometrtic_configuration.stacking_rows

    if int(parallel_e*series_e) != int(normal_count*parallel_count):
        raise Exception('Number of cells in gemetric layout not equal to number of cells in electric circuit configuration ')
         
    normal_spacing     = fuel_cell_stack.geometrtic_configuration.normal_spacing   
    parallel_spacing   = fuel_cell_stack.geometrtic_configuration.parallel_spacing
    volume_factor      = fuel_cell_stack.volume_packaging_factor 
    euler_angles       = fuel_cell_stack.orientation_euler_angles
    fuel_cell_length   = fuel_cell_stack.fuel_cell.length 
    fuel_cell_width    = fuel_cell_stack.fuel_cell.width   
    fuel_cell_height   = fuel_cell_stack.fuel_cell.height    
    
    x1 =  normal_count * (fuel_cell_length +  normal_spacing) * volume_factor # distance in the module-level normal direction
    x2 =  parallel_count *  (fuel_cell_width +parallel_spacing) * volume_factor # distance in the module-level parallel direction
    x3 =  fuel_cell_height * volume_factor # distance in the module-level height direction 

    length = x1 / stacking_rows
    width  = x2
    height = x3 *stacking_rows     
    
    if  euler_angles[0] == (np.pi / 2):
        x1prime      = x2
        x2prime      = -x1
        x3prime      = x3 
    if euler_angles[1] == (np.pi / 2):
        x1primeprime = -x3prime
        x2primeprime = x2prime
        x3primeprime = x1prime
    if euler_angles[2] == (np.pi / 2):
        length       = x1primeprime
        width        = x3primeprime
        height       = -x2primeprime

    # store length, width and height
    fuel_cell_stack.length = length
    fuel_cell_stack.width  = width
    fuel_cell_stack.height = height 
    fuel_cell              = fuel_cell_stack.fuel_cell 
     
    if type(fuel_cell_stack) == RCAIDE.Library.Components.Powertrain.Converters.Generic_Fuel_Cell_Stack:
        lb                                     = 0.0001/(Units.cm**2.)    #lower bound on fuel cell current density
        ub                                     = 1.2/(Units.cm**2.)
        sign                                   = -1. # used to minimize -power 
        maximum_current_density                = sp.optimize.fminbound(compute_power, lb, ub, args=(fuel_cell, sign)) 
        P_fuel_cell                            = compute_power(maximum_current_density,fuel_cell)  
        V_fuel_cell                            = compute_voltage(fuel_cell,maximum_current_density)  # useful voltage vector 
        efficiency                             = np.divide(V_fuel_cell, fuel_cell.ideal_voltage)
        mdot_H2                                = np.divide(P_fuel_cell,np.multiply(fuel_cell.propellant.specific_energy,efficiency))
        
        # store properties
        fuel_cell.rated_current_density       = maximum_current_density
        fuel_cell.volume                       = fuel_cell.interface_area*fuel_cell.wall_thickness
        fuel_cell.mass                         = fuel_cell.volume*fuel_cell.cell_density*fuel_cell.porosity_coefficient  
        fuel_cell.density                      = fuel_cell.mass/fuel_cell.volume                      
        fuel_cell.specific_power               = fuel_cell.max_power/fuel_cell.mass  
  
        fuel_cell_stack.mass_properties.mass   = n_total*fuel_cell.mass 
        fuel_cell_stack.voltage                = V_fuel_cell  * series_e
        fuel_cell_stack.maximum_voltage        = V_fuel_cell  * series_e
        fuel_cell_stack.maximum_power          = P_fuel_cell * series_e 
        fuel_cell_stack.maximum_current        = fuel_cell_stack.maximum_power / fuel_cell_stack.maximum_voltage
        fuel_cell_stack.maximum_fuel_flow_rate = mdot_H2 * n_total
    
    elif type(fuel_cell_stack) == RCAIDE.Library.Components.Powertrain.Converters.Proton_Exchange_Membrane_Fuel_Cell: 
    
        # check if mach number and temperature are passed
        design_altitude =  0
        
        # call the atmospheric model to get the conditions at the specified altitude
        atmosphere = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
        atmo_data  = atmosphere.compute_values(design_altitude) 
         
        segment                                     = RCAIDE.Framework.Mission.Segments.Segment()
        segment.hybrid_power_split_ratio            = 1.0
        segment.battery_fuel_cell_power_split_ratio = 0.1
        segment.state.conditions                    = RCAIDE.Framework.Mission.Common.Results()   
        bus                                         = RCAIDE.Library.Components.Powertrain.Distributors.Electrical_Bus() 
        bus.fuel_cell_stacks.append(fuel_cell_stack)   
        
        bus.append_operating_conditions(segment)
        for fuel_cell_stack in  bus.fuel_cell_stacks: 
            fuel_cell_stack.append_operating_conditions(segment,bus)      
            
        for tag, bus_item in  bus.items():  
            if issubclass(type(bus_item), RCAIDE.Library.Components.Component):
                bus_item.append_operating_conditions(segment,bus)
 
        for cryogenic_tank in  bus.cryogenic_tanks: 
            cryogenic_tank.append_operating_conditions(segment,bus)
      
        # compute fuel cell performance             
        t_idx                                                                    =  0
        fuel_cell_stack_conditions                                               = segment.state.conditions.energy[bus.tag].fuel_cell_stacks[fuel_cell_stack.tag]
        fuel_cell_stack_conditions.fuel_cell.stagnation_temperature[t_idx, 0]    = atmo_data.temperature   
        fuel_cell_stack_conditions.fuel_cell.stagnation_pressure[t_idx, 0]       = atmo_data.pressure   
        fuel_cell_stack_conditions.fuel_cell.pressure_drop[t_idx, 0]             = fuel_cell.rated_p_drop_fc
        fuel_cell_stack_conditions.fuel_cell.stack_temperature[t_idx, 0]         = fuel_cell.stack_temperature 
        rated_current_density, rated_power_density                               = evaluate_max_gross_power(fuel_cell_stack,fuel_cell_stack_conditions,t_idx)
        set_rated_current_density(fuel_cell_stack, rated_current_density, rated_power_density)
        
        fuel_cell_stack_conditions.fuel_cell.current_density[t_idx] = rated_current_density
        m_dot_H2, V_fuel_cell, P_fuel_cell, _, _, _, _, _,_  =  evaluate_PEM(fuel_cell_stack,fuel_cell_stack_conditions, t_idx)
   
        # store properties
        area_square_meters                     = fuel_cell.interface_area * 0.0001
        fuel_cell.volume                       = area_square_meters*fuel_cell.wall_thickness
        fuel_cell.mass                         = fuel_cell.volume*fuel_cell.cell_density*fuel_cell.porosity_coefficient  
        fuel_cell.density                      = fuel_cell.mass/fuel_cell.volume                      
        fuel_cell.specific_power               = fuel_cell.max_power/fuel_cell.mass  
        fuel_cell_stack.mass_properties.mass   = n_total*fuel_cell.mass 
        fuel_cell_stack.voltage                = V_fuel_cell  * series_e
        fuel_cell_stack.maximum_voltage        = V_fuel_cell  * series_e
        fuel_cell_stack.maximum_power          = P_fuel_cell * n_total 
        fuel_cell_stack.maximum_current        = fuel_cell_stack.maximum_power / fuel_cell_stack.maximum_voltage
        fuel_cell_stack.maximum_fuel_flow_rate = m_dot_H2 * n_total         
         
    return 


"""
Methods for rotor performance analysis using Blade Element Momentum Theory with Helmholtz wake modeling.

This module provides functions for analyzing rotor performance using an advanced implementation
of Blade Element Momentum Theory (BEMT) that incorporates Helmholtz wake modeling. This approach
combines the computational efficiency of BEMT with a more physically accurate representation of
the wake structure using Helmholtz vortex filaments.

The Helmholtz wake model accounts for the helical vortex structure shed from the rotor blades,
providing improved predictions of induced velocities, especially in forward flight and for
highly loaded rotors. The module includes methods for computing wake-induced velocities,
wake contraction, and overall rotor performance metrics such as thrust, torque, and power.

See Also
--------
RCAIDE.Library.Methods.Powertrain.Converters.Rotor.Performance.Actuator_Disc_Theory
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from .BEMT_Helmholtz_performance               import BEMT_Helmholtz_performance
from .compute_wake_induced_velocity            import compute_wake_induced_velocity 
from .compute_wake_contraction_matrix          import compute_wake_contraction_matrix 
from .wake_model                               import *



# RCAIDE/Library/Plots/Thermal_Management/plot_cross_flow_heat_exchanger_conditions.py
# 
# 
# Created:  Sep 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#   plot_heat_exchanger_system_conditions
# ----------------------------------------------------------------------------------------------------------------------   
def plot_cross_flow_heat_exchanger_conditions(cross_flow_hex, results, coolant_line,
                                            save_figure = False,
                                            show_legend = True,
                                            save_filename = "Cross_Flow_HEX_Conditions",
                                            file_type = ".png",
                                            width = 11, height = 7):
    """
    Creates a multi-panel visualization of cross-flow heat exchanger operating conditions.

    Parameters
    ----------
    cross_flow_hex : Component
        Cross-flow heat exchanger component containing:
            - tag : str
                Unique identifier for the heat exchanger
            
    results : Results
        RCAIDE results data structure containing:
            - segments[i].conditions.energy[coolant_line.tag][cross_flow_hex.tag]
                Heat exchanger data containing:
                    - coolant_mass_flow_rate[:,0]
                        Coolant flow rate in kg/s
                    - effectiveness_HEX[:,0]
                        Heat exchanger effectiveness
                    - power[:,0]
                        Heat transfer rate in watts
                    - air_inlet_pressure[:,0]
                        Air-side inlet pressure in Pa
                    - inlet_air_temperature[:,0]
                        Air inlet temperature in K
                    - air_mass_flow_rate[:,0]
                        Air flow rate in kg/s
                    
    coolant_line : Component
        Coolant line component containing:
            - tag : str
                Unique identifier for the coolant circuit
            
    save_figure : bool, optional
        Flag for saving the figure (default: False)
        
    show_legend : bool, optional
        Flag to display segment legend (default: True)
        
    save_filename : str, optional
        Name of file for saved figure (default: "Cross_Flow_HEX_Conditions")
        
    file_type : str, optional
        File extension for saved figure (default: ".png")
        
    width : float, optional
        Figure width in inches (default: 11)
        
    height : float, optional
        Figure height in inches (default: 7)

    Returns
    -------
    fig : matplotlib.figure.Figure
        Handle to the generated figure containing six subplots arranged in a 3x2 grid:
        
    Notes
    -----
    Creates visualization showing:
        * Thermal performance metrics
        * Flow conditions for both fluids
        * Heat transfer characteristics
        * Operating pressures and temperatures

    **Definitions**
    
    'Effectiveness'
        Ratio of actual to maximum possible heat transfer
    'NTU'
        Dimensionless measure of heat exchanger size
    'Capacity Ratio'
        Ratio of minimum to maximum heat capacity rates
    'Mass Flow Rate'
        Mass of fluid flowing per unit time
    
    See Also
    --------
    RCAIDE.Library.Plots.Thermal_Management.plot_thermal_management_performance : Overall system performance
    RCAIDE.Library.Plots.Thermal_Management.plot_air_cooled_conditions : Air-cooled heat exchanger analysis
    """ 
    
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))     

    fig = plt.figure(save_filename)
    fig.set_size_inches(width,height)  
    axis_1 = plt.subplot(3,2,1)
    axis_2 = plt.subplot(3,2,2) 
    axis_3 = plt.subplot(3,2,3) 
    axis_4 = plt.subplot(3,2,4)
    axis_5 = plt.subplot(3,2,5) 
    axis_6 = plt.subplot(3,2,6) 
 
    for network in results.segments[0].analyses.energy.vehicle.networks: 
        busses  = network.busses 
        for bus in busses:
            for b_i, battery in enumerate(bus.battery_modules):
                if b_i == 0 or bus.identical_battery_modules == False:
                    for i in range(len(results.segments)): 
                        time    = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min    
                        cross_flow_hex_conditions  = results.segments[i].conditions.energy[coolant_line.tag][cross_flow_hex.tag]  
                
                        coolant_mass_flow_rate     = cross_flow_hex_conditions.coolant_mass_flow_rate[:,0]        
                        effectiveness_HEX          = cross_flow_hex_conditions.effectiveness_HEX[:,0]   
                        power                      = cross_flow_hex_conditions.power[:,0]                       
                        inlet_air_pressure         = cross_flow_hex_conditions.air_inlet_pressure[:,0]          
                        inlet_air_temperature      = cross_flow_hex_conditions.inlet_air_temperature[:,0]          
                        air_mass_flow_rate         = cross_flow_hex_conditions.air_mass_flow_rate[:,0]     
                                             
                        if i == 0: 
                            axis_1.plot(time, effectiveness_HEX, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width, label = cross_flow_hex.tag)
                        else:
                            axis_1.plot(time, effectiveness_HEX, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width) 
                        axis_1.set_ylabel(r'Effectiveness') 
                        set_axes(axis_1)      
                
                        axis_2.plot(time,  inlet_air_temperature, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_2.set_ylabel(r'Air Temp. (K)') 
                        set_axes(axis_2)    
                        
                        axis_3.plot(time, coolant_mass_flow_rate, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_3.set_ylabel(r'Coolant $\dot{m}$ (kg/s)')
                        set_axes(axis_3) 
                
                        axis_4.plot(time, air_mass_flow_rate, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_4.set_ylabel(r'Air $\dot{m}$ (kg/s)')
                        set_axes(axis_4)                               
                 
                        axis_5.plot(time, power/1000, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_5.set_ylabel(r'HEX Power (KW)')
                        axis_5.set_xlabel(r'Time (mins)')
                        set_axes(axis_5)    
                
                        axis_6.plot(time, inlet_air_pressure/10e6 , color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_6.set_ylabel(r'Air Pres. (MPa)')
                        axis_6.set_xlabel(r'Time (mins)')
                        set_axes(axis_6)  
            
    if show_legend:     
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4) 
        leg.set_title('Flight Segment', prop={'size': ps.legend_font_size, 'weight': 'heavy'})  
    
    # Adjusting the sub-plots for legend 
    fig.tight_layout()
    fig.subplots_adjust(top=0.8) 
    
    # set title of plot 
    title_text   = 'Heat_Exchanger_System'       
    fig.suptitle(title_text) 
    
    if save_figure:
        plt.savefig(save_filename + cross_flow_hex.tag + file_type)    
    return fig 


# RCAIDE/Library/Methods/Powertrain/Converters/Motor/__init__.py

# Created:  Jan 2025, M. Clarke, M. Guidotti

"""
Motor Methods Package

This module contains methods for motor performance analysis, design, and condition handling.
The methods support both DC and PMSM motor modeling within the RCAIDE framework.

See Also
--------
RCAIDE.Library.Components.Powertrain.Converters.DC_Motor
RCAIDE.Library.Components.Powertrain.Converters.PMSM_Motor
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .compute_motor_performance             import compute_motor_performance
from .design_optimal_motor                  import design_optimal_motor
from .append_motor_conditions               import append_motor_conditions 


# RCAIDE/Library/Attributes/Solids/Polyetherimide.py
# 
# 
#
# Created: Sep 2024 S. Shekar

#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------

from .Solid import Solid
from RCAIDE.Framework.Core import Units

#-------------------------------------------------------------------------------
# Polyetherimide for Reservoir Casing
#-------------------------------------------------------------------------------
class Polyetherimide(Solid):

    """
    A class representing polyetherimide (PEI) material properties, commonly used in reservoir casings 
    and other high-temperature thermoplastic applications.

    Attributes
    ----------
    conductivity : float
        Thermal conductivity in W/(m·K) (0.2)
    
    emissivity : float
        Surface emissivity, unitless (0.96)
    
    specific_heat : float
        Specific heat capacity in J/(kg·K) (1100)

    Notes
    -----
    This class implements thermal properties for polyetherimide, a high-performance 
    thermoplastic known for its heat resistance and dimensional stability. The material 
    is commonly used in aerospace and high-temperature applications.

    **Definitions**
    
    'Thermal Conductivity'
        The property of a material to conduct heat, measured in watts per meter-kelvin
    
    'Emissivity'
        The effectiveness of a material's surface in emitting energy as thermal radiation
    
    'Specific Heat'
        The amount of heat required to raise the temperature of 1 kg of the material by 1 Kelvin

    References
    ----------
    [1] MatWeb. (n.d.-b).  Overview of materials for Polyetherimide (PEI). MatWeb Material Property Data. https://www.matweb.com/search/DataSheet.aspx?MatGUID=65baf7a4f90c4a54a6ace03e16b1125b&amp%3Bckck=1&ckck=1 
    """
    def __defaults__(self):
        """Sets material properties at instantiation.

        Assumptions:
        None

        Source:
        https://www.matweb.com/search/DataSheet.aspx?MatGUID=65baf7a4f90c4a54a6ace03e16b1125b&amp%3bckck=1&ckck=1

        Inputs:
        N/A

        Outputs:
        N/A

        Properties Used:
        None
        """ 
        self.conductivity              = 0.2    # [W/m-K]
        self.emissivity                = 0.96   # [uniteless]
        self.specific_heat             = 1100  



# RCAIDE/Library/Missions/Segments/Cruise/Constant_Acceleration_Constant_Altitude.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# Package imports 
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def initialize_conditions(segment):
    """
    Initializes conditions for constant acceleration cruise at fixed altitude

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed

    Notes
    -----
    This function sets up the initial conditions for a cruise segment with constant
    acceleration and constant altitude. The velocity changes linearly with time
    based on the specified acceleration.

    **Required Segment Components**

    segment:
        - altitude : float
            Cruise altitude [m]. If not specified, the altitude from the previous segment is used.
        - air_speed_start : float
            Initial true airspeed [m/s]. If not specified, the airspeed from the previous segment is used.
        - air_speed_end : float
            Final true airspeed [m/s]
        - acceleration : float
            Constant acceleration [m/s^2]
        - sideslip_angle : float
            Aircraft sideslip angle [rad]
        - state:
            numerics.dimensionless.control_points : array
                Discretization points [-]
            conditions : Data
                State conditions container

    **Calculation Process**
        1. Calculate time required for speed change
        2. Discretize time points
        3. Calculate velocity profile using:
            - Initial velocity
            - Constant acceleration
            - Sideslip angle for lateral components

    **Major Assumptions**
        * Constant acceleration
        * Constant altitude
        * Small angle approximations
        * Quasi-steady flight
        * No wind effects

    Returns
    -------
    None
        Updates segment conditions directly:
        - conditions.frames.inertial.velocity_vector [m/s]
        - conditions.frames.inertial.position_vector [m]
        - conditions.freestream.altitude [m]
        - conditions.frames.inertial.time [s]

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    """      
    
    # unpack
    alt        = segment.altitude 
    v0         = segment.air_speed_start
    vf         = segment.air_speed_end
    ax         = segment.acceleration    
    beta       = segment.sideslip_angle
    conditions = segment.state.conditions 
    
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
    
    # check for initial velocity
    if v0 is None: 
        if not segment.state.initials: raise AttributeError('airspeed not set')
        v0 = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])
        
    # dimensionalize time
    t_initial = conditions.frames.inertial.time[0,0]
    t_final   = (vf-v0)/ax + t_initial
    t_nondim  = segment.state.numerics.dimensionless.control_points
    time      = t_nondim * (t_final-t_initial) + t_initial 
    v_mag = v0+(time - t_initial)*ax
    v_x   = np.cos(beta)*v_mag
    v_y   = np.sin(beta)*v_mag
    
    # pack
    segment.state.conditions.freestream.altitude[:,0]             = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down
    segment.state.conditions.frames.inertial.velocity_vector[:,0] = v_x[:,0]
    segment.state.conditions.frames.inertial.velocity_vector[:,1] = v_y[:,0]
    segment.state.conditions.frames.inertial.time[:,0]            = time[:,0] 
